/*******************************************************************************
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package javax.tools.diagnostics.vm.spi.delegates;

import java.io.File;
import java.io.IOException;
import java.lang.management.ManagementFactory;

import javax.management.InstanceNotFoundException;
import javax.management.MBeanException;
import javax.management.MBeanServer;
import javax.management.MalformedObjectNameException;
import javax.management.ObjectName;
import javax.management.ReflectionException;
import javax.tools.diagnostics.vm.DumpDescriptor;
import javax.tools.diagnostics.vm.DumpHandle;
import javax.tools.diagnostics.vm.DumpInitiatorDelegate;
import javax.tools.diagnostics.vm.spi.DumpInitiatorCapabilities;

/**
 * Uses HotSpotDiagnostic MBean to generate hprof dump.
 * Only available on Sun's hotspot.
 *
 */
public class HProfMBeanDumpDelegate implements DumpInitiatorDelegate {
	private MBeanServer mbeanServer;
	private ObjectName objName;
	private MBeanDumpHandle handle = null;
	
	public HProfMBeanDumpDelegate() {
		mbeanServer = ManagementFactory.getPlatformMBeanServer();
		try {
			objName = ObjectName.getInstance("com.sun.management:type=HotSpotDiagnostic");
		} catch (MalformedObjectNameException e) {
			// Should be ok.
		}
		
		if (available()) {
			handle = new MBeanDumpHandle();
		}
	}

	/**
	 * Checks to see if the MBean is registered.
	 * On IBM JVMs, this will return false.
	 */
	@Override
	public boolean available() {
		return mbeanServer.isRegistered(objName);
	}

	@Override
	public DumpHandle createDumpHandle() {
		return handle;
	}

	@Override
	public DumpHandle createDumpHandle(DumpDescriptor descriptor) {
		// TODO Ignores DumpDescriptor - we have no control over contents.
		return handle;
	}

	@Override
	public DumpInitiatorCapabilities getCapabilities() {
		// TODO Will need real implementation.
		return new DumpInitiatorCapabilities() {
		};
	}

	@Override
	public String getDumpType() {
		return "sun.hprof.mbean";
	}

	private class MBeanDumpHandle implements DumpHandle {

		/**
		 * Initiates an hprof dump.
		 * Generates a file with a name:
		 * 	hprofDump.time in milliseconds.hprof
		 * 
		 * This is configurable - we just need the interface to do so.
		 */
		@Override
		public boolean dump() throws IOException {
			String fileName = System.getProperty("user.dir")
			+ File.separatorChar + "hprofDump."
			+ System.currentTimeMillis() + ".hprof";



			if (available()) {
				try {
					
					// Parameters are the file name, and a boolean that
					// should be true if only live objects are to be dumped.
					mbeanServer.invoke(objName,
							"dumpHeap", new Object[] { fileName, true }, new String[] {
							"java.lang.String", "boolean" });
					return true;
				} catch (InstanceNotFoundException e) {
					e.printStackTrace();
				} catch (MBeanException e) {
					e.printStackTrace();
				} catch (ReflectionException e) {
					e.printStackTrace();
				}
			}
			return false;
		}

	}
}
