/*******************************************************************************
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package javax.tools.diagnostics.image;

/**
 * Represents an address in image memory.
 * 
 */
public interface ImagePointer {

    /**
     * Get the unwrapped address, represented as a 64-bit integer.
     * @return the unwrapped address, represented as a 64-bit integer
     * <br>
     * Use caution when comparing addresses, as some addresses may be
     * negative.
     * <br>
     * Note that on segmented memory architectures, it may not be 
     * possible to represent all addresses accurately as integers
     */
    long getAddress();
    
    /**
     * Get the address space to which this pointer belongs.
     * @return the address space to which this pointer belongs
     */
    ImageAddressSpace getAddressSpace();
    
    /**
     * Build a new image pointer offset from this one by the given amount.
     * @param offset
     * @return a new ImagePointer based at getAddress() + offset
     */
    ImagePointer add(long offset);
    
    /**
     * Is the referenced location executable ?
     * @return true if this memory address is within an executable page
     * @throws DataUnavailable 
     */
    boolean isExecutable() throws DataUnavailable;
    
    /**
     * Is the referenced location read only ?
     * @return true if write access to this memory address was disabled in the image
     * @throws DataUnavailable 
     */
    boolean isReadOnly() throws DataUnavailable;
    
    /**
     * Is the referenced location shared ?
     * @return true if this memory address is shared between processes
     * @throws DataUnavailable 
     */
    boolean isShared() throws DataUnavailable;
    
    /**
     * Get the value at the given offset from this pointer.
     * To determine the number of bytes to skip after this call to read the next value, use <code>ImageProcess.getPointerSize()</code>.
     * @param index  an offset (in bytes) from the current position
     * @return the 32 or 64-bit pointer stored at getAddress() + index in the same address space.
     * @throws MemoryAccessException if the memory cannot be read
     * @throws CorruptDataException if the memory should be in the image, but is missing or corrupted
     * @see    ImageProcess#getPointerSize()
     */
    ImagePointer getPointerAt(long index) throws MemoryAccessException, CorruptDataException;
    
    /**
     * Get the value at the given offset from this pointer.
     * @param index  an offset (in bytes) from the current position
     * @return the 64-bit long stored at getAddress() + index 
     * @throws MemoryAccessException if the memory cannot be read
     * @throws CorruptDataException if the memory should be in the image, but is missing or corrupted
     */
    long getLongAt(long index)  throws MemoryAccessException, CorruptDataException;
    
    /**
     * Get the value at the given offset from this pointer.
     * @param index  an offset (in bytes) from the current position
     * @return the 32-bit int stored at getAddress() + index 
     * @throws MemoryAccessException if the memory cannot be read
     * @throws CorruptDataException if the memory should be in the image, but is missing or corrupted
     */
    int getIntAt(long index) throws MemoryAccessException, CorruptDataException;
    
    /**
     * Get the value at the given offset from this pointer.
     * @param index  an offset (in bytes) from the current position
     * @return the 16-bit short stored at getAddress() + index 
     * @throws MemoryAccessException if the memory cannot be read
     * @throws CorruptDataException if the memory should be in the image, but is missing or corrupted
     */
    short getShortAt(long index) throws MemoryAccessException, CorruptDataException;
    
    /**
     * Get the value at the given offset from this pointer.
     * @param index  an offset (in bytes) from the current position
     * @return the 8-bit byte stored at getAddress() + index 
     * @throws MemoryAccessException if the memory cannot be read
     * @throws CorruptDataException if the memory should be in the image, but is missing or corrupted
     */
    byte getByteAt(long index) throws MemoryAccessException, CorruptDataException;
    
    /**
     * Get the value at the given offset from this pointer.
     * @param index  an offset (in bytes) from the current position
     * @return the 32-bit float stored at getAddress() + index
     * @throws MemoryAccessException if the memory cannot be read
     * @throws CorruptDataException if the memory should be in the image, but is missing or corrupted
     */
    float getFloatAt(long index) throws MemoryAccessException, CorruptDataException;

    /**
     * Get the value at the given offset from this pointer.
     * @param index  an offset (in bytes) from the current position
     * @return the 64-bit double stored at getAddress() + index 
     * @throws MemoryAccessException if the memory cannot be read
     * @throws CorruptDataException if the memory should be in the image, but is missing or corrupted
     */
    double getDoubleAt(long index) throws MemoryAccessException, CorruptDataException;
    
	/**
	 * @param obj
	 * @return True obj refers to the same Image Pointer in the image
	 */
	public boolean equals(Object obj);
	public int hashCode();
}
