/*******************************************************************************
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package javax.tools.diagnostics.image;

import java.util.List;
import java.util.Properties;

import javax.tools.diagnostics.runtime.ManagedRuntime;

/**
 * This class represents a Process running in a given Address Space.
 *
 */
public interface ImageProcess {

    /**
     * Fetch the command line for this process. This may be the exact command line
     * the user issued, or it may be a reconstructed command line based on
     * argv[] and argc.
     * 
     * @return the command line for the process
     *
     * @exception DataUnavailable if the information cannot be provided
     * @throws CorruptDataException 
     */
    String getCommandLine() throws DataUnavailable, CorruptDataException;
    
    /**
     * Get the environment variables for this process.
     * @return the environment variables for this process
     * 
     * @exception DataUnavailable if the information cannot be provided
     * @throws CorruptDataException 
     */
    Properties getEnvironment() throws DataUnavailable, CorruptDataException;
    
    /**
     * Get the system-wide identifier for the process.
     * @return a system-wide identifier for the process (e.g. a process id (pid) on Unix like
     * systems)
     * 
   	 * @exception DataUnavailable if the information cannot be provided
     * @throws CorruptDataException 
     */
    String getID() throws DataUnavailable, CorruptDataException;
    
    /**
     * Get the set of shared libraries which are loaded in this process.
     * @return an iterator to iterate over the shared libraries which are loaded in this
     * process
     * 
     * @exception DataUnavailable if the information cannot be provided
     * @throws CorruptDataException 
     * 
     * @see ImageModule
     */
    List<ImageModule> getLibraries() throws DataUnavailable, CorruptDataException;
    
    /**
     * Get the module representing the executable within the image.
     * @return the module representing the executable within the image (as opposed to
     * modules representing libraries)
     * 
     * @exception DataUnavailable if the information cannot be provided
     * @throws CorruptDataException 
     * 
     * @see ImageModule
     */
    ImageModule getExecutable() throws DataUnavailable, CorruptDataException;
    
    /**
     * Get the set of image threads in the image.
     * 
     * There is not necessarily any relationship between JavaThreads and
     * ImageThreads. A JVM implementation may use an n:m mapping of JavaThreads to
     * ImageThreads, and not all ImageThreads are necessarily attached.
     * 
     * @return an iterator to iterate over each ImageThread in the image
     * @see ImageThread 
     * @see CorruptData
     */
    List<ImageThread> getThreads();

    /**
     * Find the thread which triggered the creation of the image
     * 
     * @return the ImageThread which caused the image to be created, or
     * null if the image was not created due to a specific thread
     * @throws CorruptDataException 
     * 
     * @see ImageThread
     */
    ImageThread getCurrentThread() throws CorruptDataException;
    
    /**
     * Get the set of the known ManagedRuntime environments in the image. 
     * In a typical image, there will be only one runtime, and it will be
     * an instance of JavaRuntime. However any user of this API should be aware
     * that there is a possibility that other runtimes may exist in the image
     * 
     * @return an iterator to iterate over all of the known ManagedRuntime
     * environments in the image. 
     * 
     * 
     * @see javax.tools.diagnostics.runtime.ManagedRuntime 
     * @see javax.tools.diagnostics.runtime.java.JavaRuntime
     */
    List<ManagedRuntime> getRuntimes();
    
    /**
     * Get the OS signal number in this process which triggered the creation 
     * of this image.
     * 
     * @return the OS signal number in this process which triggered the creation 
     * of this image, or 0 if the image was not created because of a signal in this
     * process
     * 
     * @exception DataUnavailable if the information cannot be provided
     * @throws CorruptDataException 
     */
    int getSignalNumber() throws DataUnavailable, CorruptDataException;
    
    /**
     * Get the name of the OS signal in this process which triggered the 
     * creation of this image.
     * @return the name of the OS signal in this process which triggered the 
     * creation of this image, or null if the image was not created because of a 
     * signal in this process
     * 
     * @exception DataUnavailable if the information cannot be provided
     * @throws CorruptDataException 
     */
    String getSignalName() throws DataUnavailable, CorruptDataException;
    
    /**
     * Determine the pointer size used by this process.
     * Currently supported values are 31, 32 or 64.
     * In the future, other pointer sizes may also be supported.
     * 
     * @return the size of a pointer, in bits
     */
    int getPointerSize();

}
