/*******************************************************************************
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package javax.tools.diagnostics.image;

/**
 * Represents a native stack frame
 */
public interface ImageStackFrame {

    /**
     * Get the address of the current instruction within
     * the procedure being executed.
     * @return the address of the current instruction within
     * the procedure being executed, or null if not available.
     * <br>
     * Use this address with caution, as it is provided only 
     * as a best guess. It may not be correct, or even within
     * readable memory
     * @throws CorruptDataException 
     */
    ImagePointer getProcedureAddress() throws CorruptDataException;
    
    /**
     * Get the base pointer of the stack frame.
     * @return the base pointer of the stack frame
     * @throws CorruptDataException 
     */
    ImagePointer getBasePointer() throws CorruptDataException;
    
    /**
     * Returns a string describing the procedure at this stack
     * frame. Implementations should use the following template
     * so that procedure names are reported consistently:
     * <p>
     * <dl><dd><pre>libname(sourcefile)::entrypoint&#177;<!-- +/- -->offset</pre></dd></dl>
     * <p>
     * Any portion of the template may be omitted if it is not available 
     * <dl>
     * <dt>e.g.</dt>
     * 		<dd><pre>system32(source.c)::WaitForSingleObject+14</pre></dd>
     * 		<dd><pre>system32::WaitForSingleObject-4</pre></dd>
     * 		<dd><pre>(source.c)::WaitForSingleObject</pre></dd>
     * 		<dd><pre>::WaitForSingleObject+14</pre></dd>
     * 		<dd><pre>system32+1404</pre></dd>
     * 		<dd><pre>system32::TWindow::open(int,void*)+14</pre></dd>
     * </dl>
     * 
     * @return a string naming the function executing in this
     * stack frame.  If the name is not known for legitimate
     * reasons, a synthetic name will be returned..
     * @throws CorruptDataException 
     */
    String getProcedureName() throws CorruptDataException;
    
}
