/*******************************************************************************
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package javax.tools.diagnostics.image;

import java.util.List;
import java.util.Properties;

/**
 * A low-level thread instance
 */
public interface ImageThread {

    /**
     * Fetch a unique identifier for the thread.
     * In many operating systems, threads have more than one identifier (e.g.
     * a thread id, a handle, a pointer to VM structures associated with the thread).
     * In this case, one of these identifiers will be chosen as the canonical
     * one. The other identifiers would be returned by getProperties()
     * 
     * @return a process-wide identifier for the thread (e.g. a tid number)
     * @throws CorruptDataException 
     */
    String getID() throws CorruptDataException;
 
    /**
     * Get the set of stack frames on this thread.
     * 
     * @return an iterator to walk the native stack frames in order from 
     * top-of-stack (that is, the most recent frame) to bottom-of-stack. Throws
     * DataUnavailable if native stack frames are not available on this platform.
     * 
     * @throws DataUnavailable If native stack frames are not available on this platform
     * @see ImageStackFrame
     * 
     */
    List<ImageStackFrame> getStackFrames() throws DataUnavailable;
    
    /**
     * Get the set of image sections which make up the stack.
     * 
     * @return a collection of ImageSections which make up the stack. On
     * most platforms this consists of a single entry, but on some platforms
     * the thread's stack may consist of non-contiguous sections
     * 
     * @see ImageSection
     * @see CorruptData
     */
    List<ImageSection> getStackSections();
    
    /**
     * Get the register contents.
     * 
     * @return an iterator to iterate over the state of the CPU registers
     * when the image was created. The collection may be empty if the register 
     * state is not available for this thread.
     * 
     * If the CPU supports partial registers (e.g. AH, AL, AX, EAX, RAX on
     * AMD64), only the largest version of the register will be included
     * 
     * @see ImageRegister
     */
    List<ImageRegister> getRegisters();
    
    /**
     * Get the OS-specific properties for this thread.
     * @return a table of OS-specific properties for this thread.
     * Values which are commonly available include
     * <ul>
     * <li>"priority" -- the priority of the thread</li>
     * <li>"policy" -- the scheduling policy of the thread</li>
     * </ul> 
     */
    Properties getProperties();
}
