/*******************************************************************************
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package javax.tools.diagnostics.runtime.java;

import java.util.List;

import javax.tools.diagnostics.image.ImageSection;

/**
 * Represents a single heap of managed objects.
 * 
 * The heap can be viewed as an unordered collection of JavaObjects or as a region of storage 
 * within the Java Virtual Machine instance. 
 * 
 * The heap commonly contains JavaObject instances that are reachable by navigating chains of {@link JavaReference}
 * These references can be obtained from the {@link JavaRuntime#getHeapRoots()} method.
 * 
 * A heap can contain instances which cannot be reached by the use of {@link JavaReference}  
 * 
 */
public interface JavaHeap {
    
    /**
     * Get the set of memory regions that represent the memory layout of the heap.
     * The actual make up of this list is implementation specific.
     * 
     * The returned list follows the standard semantics for javax.tools.diagnostics collections.
     * 
     * The returned value is never null but can be an empty list.
     *
     * @return a list  of {@link ImageSection} instances 
     * 
     * 
     * @see javax.tools.diagnostics.image.ImageSection
     */
    List<ImageSection> getSections();
    
    /**
     * Get a brief textual description of this heap.
     * The value returned is  implementation specific.
     * 
     * The returned value is never null.
     * 
     * @return a brief textual description of this heap
     */
    String getName();
    
    /**
     * Get the set of objects which are stored in this heap.
     * 
     * @return a list  of {@link JavaObject} objects which are stored in this heap
     * 
     * The returned list follows the standard semantics for javax.tools.diagnostics collections.
     * 
     * The returned value is never null but can be an empty list.
     * 
     * @see JavaObject
     */
    List<JavaObject> getObjects();
    
	/**
	 * @param obj
	 * @return true if the given object refers to the same Java Heap in the image
	 */
	public boolean equals(Object obj);
	public int hashCode();
}
