/*******************************************************************************
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package javax.tools.diagnostics.runtime.java;

import javax.tools.diagnostics.image.CorruptDataException;
import javax.tools.diagnostics.image.DataUnavailable;
import javax.tools.diagnostics.image.ImagePointer;

/**
 * <p>
 * Represents a point of execution within a Java method
 * 
 */
public interface JavaLocation {
	
	/**
	 * <p>
	 * Fetches the absolute address of the code which this location represents.
	 * This pointer will be contained within one of the segments returned by
	 * getBytecodeSections() or getCompiledSections() of the method returned by
	 * getMethod().
	 * <p>
	 * null may be returned, particularly for methods with no bytecode or 
	 * compiled sections (e.g. some native methods)
	 * <p>
	 * Although an offset into the method may be calculated using this 
	 * pointer, caution should be exercised in attempting to map this offset to
	 * an offset within the original class file. Various transformations may
	 * have been applied to the bytecodes by the VM or other agents which may
	 * make the offset difficult to interpret.
	 * <p>
	 * For native methods, the address may be meaningless.
	 * 
	 * @return the address in memory of the managed code
	 * @throws CorruptDataException if the underlying data is in an unexpected state 
	 */
	ImagePointer getAddress() throws CorruptDataException;
	
	/**
	 * <p>
	 * Get the line number.
	 * <p>
	 * @return the line number, if available, or throws DataUnavailable if it is not available
	 * Line numbers are counted from 1  
	 * 
	 * @throws DataUnavailable if the line number data is not available for this location
	 * @throws CorruptDataException if the underlying data is in an unexpected state 
	 */
	int getLineNumber() throws DataUnavailable, CorruptDataException;
	
	/**
	 * <p>
	 * Get the source file name.
	 * 
	 * @return the name of the source file, if available, or throws DataUnavailable if it is
	 * not available
	 * 
	 * @throws DataUnavailable if the source file name is unavailable in the core
	 * @throws CorruptDataException  if the underlying data is in an unexpected state 
	 */
	String getFilename() throws DataUnavailable, CorruptDataException;
	
	/**
	 * <p>
	 * Get the compilation level for this location. This is an implementation 
	 * defined number indicating the level at which the current location was 
	 * compiled. 0 indicates interpreted. Any positive number indicates some 
	 * level of JIT compilation. Typically, higher numbers indicate more 
	 * aggressive compilation strategies
	 * <p>
	 * For native methods, a non-zero compilation level indicates that some
	 * level of JIT compilation has been applied to the native call (e.g. a
	 * custom native call stub). To determine if the method is native, use
	 * getMethod().getModifiers().
	 * 
	 * @return the compilation level 
	 * @throws CorruptDataException if the underlying data is in an unexpected state  
	 */
	int getCompilationLevel() throws CorruptDataException;
	
	/**
	 * <p>
	 * Get the method which contains the point of execution.
	 * 
	 * @return the method which contains the point of execution
	 * @throws CorruptDataException if the underlying data is in an unexpected state  
	 */
	JavaMethod getMethod() throws CorruptDataException;
	
	/**
	 * @return A string representing the location as it would be seen in a Java stack trace
	 */
	public String toString();
	
	/**
	 * @param obj
	 * @return True if the given object refers to the same Java Location in the image
	 */
	public boolean equals(Object obj);
	public int hashCode();
}
