/*******************************************************************************
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package javax.tools.diagnostics.runtime.java;


import java.util.List;

import javax.tools.diagnostics.image.CorruptDataException;
import javax.tools.diagnostics.image.DataUnavailable;
import javax.tools.diagnostics.image.ImagePointer;
import javax.tools.diagnostics.image.ImageSection;
import javax.tools.diagnostics.image.ImageThread;



/**
 * <p>Represents a Java thread.</p>
 * <p><code>JavaThread</code> instances correspond with executing threads in the Java Virtual Machine,
 * not {@link java.lang.Thread} instances on the heap.
 * <code>JavaThread</code> provide information on what was running including the locations
 * of all of the threads within the Java program when the dump was taken.
 * </p> 
 *  
 * @see JavaRuntime#getThreads()
 */
public interface JavaThread  {

	/** The thread is alive	 */
	int STATE_ALIVE						= 0x00000001;
	/** The thread has terminated */
	int STATE_TERMINATED				= 0x00000002;
	/** The thread can be run although may not be actually running */
	int STATE_RUNNABLE					= 0x00000004;
	/** The thread is waiting on a monitor with no timeout value set */
	int STATE_WAITING_INDEFINITELY		= 0x00000010;
	/** The thread is waiting on a monitor but will timeout at some point */
	int STATE_WAITING_WITH_TIMEOUT		= 0x00000020;
	/** The thread is in the Thread.sleep method	 */
	int STATE_SLEEPING					= 0x00000040;
	/** The thread is in a waiting state in native code	 */
	int STATE_WAITING					= 0x00000080;
	/** The thread is in Object.wait */
	int STATE_IN_OBJECT_WAIT			= 0x00000100;
	/** The thread has been deliberately removed from scheduling */
	int STATE_PARKED					= 0x00000200;
	/** The thread is waiting to enter an object monitor	 */
	int STATE_BLOCKED_ON_MONITOR_ENTER	= 0x00000400;
	/** The thread has been suspended by Thread.suspend */
	int STATE_SUSPENDED					= 0x00100000;
	/** The thread has a pending interrupt */
	int STATE_INTERRUPTED				= 0x00200000;
	/** The thread is in native code */
	int STATE_IN_NATIVE					= 0x00400000;
	/** The thread is in a vendor specific state */
	int STATE_VENDOR_1					= 0x10000000;
	/** The thread is in a vendor specific state */
	int STATE_VENDOR_2					= 0x20000000;
	/** The thread is in a vendor specific state */
	int STATE_VENDOR_3					= 0x40000000;
	
    /**
     * <p>Get the address of the JNIEnv structure which represents this thread instance in JNI.</p>
     * 
     * @return the address of the JNIEnv structure which represents this thread instance in JNI.
     * @throws CorruptDataException 
     */
    ImagePointer getJNIEnv() throws CorruptDataException;
    
    /**
     * <p>Get the Java priority of the thread.</p>
     * <p>The value returned will be the same as what would have been returned by a call
     * to {@link java.lang.Thread#getPriority()} within the Java Virtual Machine.
     * 
     * @return the Java priority of the thread (a number between 1 and 10 inclusive)
     * @throws CorruptDataException 
     * 
     * @see java.lang.Thread#getPriority()
     */
    int getPriority() throws CorruptDataException;
    

    /**
     * <p>Returns the {@link JavaObject} representing the instance of the class or subclass of {@link java.lang.Thread}
     * that represents this thread in the Java Virtual Machine.</p>
     * 
     * <p>The object returned is the {@link java.lang.Thread} instance the method 
     * {@link java.lang.Thread#start() start()} was executed against in order to create this 
     * Java thread.
     * </p>
     * 
     * <p>This method may return <code>null</code> when there is no {@link java.lang.Thread} instance associated with this
     * Java thread. Some Java threads may be created for purposes other than for executing Java code
     * (for example, for garbage collection).    
     * </p>
     * 
     * 
     * @return a JavaObject representing the {@link java.lang.Thread} associated with this thread, or <code>null</code>.
     * @throws CorruptDataException if the reference to <code>java.lang.Thread</code> is not <code>null</code> and cannot be retrieved.
     * 
     * @see JavaObject
     * @see java.lang.Thread
     */
    JavaObject getObject() throws CorruptDataException;
    
    /**
     * <p>Get the state of the thread when the dump was generated.</p>
     * <p>The result is a bit vector, and uses the states defined by
     * the function 
     * <a href="http://java.sun.com/j2se/1.5.0/docs/guide/jvmti/jvmti.html#GetThreadState">GetThreadState</a>
     *  in the JVMTI specification.
     * </p> 
     * 
     * @return the state of the thread when the image was created.
     * 
     * @throws CorruptDataException If the thread state could not be successfully retrieved.  
     */
    int getState() throws CorruptDataException;
    
    /**
     * <p>Returns the operating system level thread that executes the Java thread.</p>
     * <p>This will return an {@link ImageThread} if an operating system level thread can be
     * returned, otherwise the <code>DataUnavailable</code> exception is thrown.
     * There is no guarantee that there is a 1:1 relationship between <code>JavaThreads</code>
     * and {@link ImageThread ImageThreads}.
     * </p>
     * 
     * @return the ImageThread which this thread is currently bound to.
     * @throws CorruptDataException If the underlying resource describing the native representation of the thread
     * is damaged.
     * @throws DataUnavailable If no mapping is provided due to missing or limited underlying resources.
     * 
     * @see ImageThread
     */
    ImageThread getImageThread() throws CorruptDataException, DataUnavailable;

    /**
     * <p>Get the List of {@link ImageSection ImageSections} which make up the Java Virtual Machine stack.
     * </p>
     * <p> 
     * Some Java Virtual Machine implementations may use parts of the ImageThread's stack
     * for JavaStackFrames.
     * </p>
     * 
     * @return a collection of <code>ImageSections</code> which make up the Java stack.
     * 
     * @see javax.tools.diagnostics.image.ImageSection
     * @see javax.tools.diagnostics.image.ImageThread#getStackSections()
     * @see javax.tools.diagnostics.image.CorruptData
     */
    List<ImageSection> getStackSections();
    
    /**
     * <p>Get the set of stack frames.</p>
     * <p>The start of the list will contain the top most stack frame, the last entry
     * will contain the bottom most stack frame. The top contains the most recently executing
     * stack frame.  
     * </p>
     * <p>
     * This method may return an empty list when there are no Java stack frames associated with
     * this Java thread. <code>null</code> must never be returned.
     * </p>
     * 
     * @return a list of Java stack frames in order from top to bottom.
     *  
     * @see JavaStackFrame
     * 
     */
    List<JavaStackFrame> getStackFrames();
    
    /**
     * <p>Return the name of the thread.</p>
     * 
     * <p>Usually this is derived from the object associated with the thread, but if the
     * name cannot be derived this way (e.g. there is no object associated with the thread)
     * a name will be created for the thread.
     * </p>
     * 
     * @return the name of the thread
     * @throws CorruptDataException If a name exists but cannot be retrieved. 
     */
    String getName() throws CorruptDataException;
    
	/**
	 * @param obj
	 * @return true if the given object refers to the same Java Thread in the image
	 */
	public boolean equals(Object obj);
	public int hashCode();
}
