/*******************************************************************************
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/

package org.apache.kato.jvmti.javaruntime.model;

import java.util.logging.Level;

import javax.tools.diagnostics.image.CorruptDataException;
import javax.tools.diagnostics.image.DataUnavailable;
import javax.tools.diagnostics.image.MemoryAccessException;
import javax.tools.diagnostics.runtime.java.JavaClass;
import javax.tools.diagnostics.runtime.java.JavaField;
import javax.tools.diagnostics.runtime.java.JavaObject;

import org.apache.kato.jvmti.process.CorruptDataImpl;
import org.apache.kato.jvmti.reader.CLogger;

public class JField implements JavaField, JValueHolder {

	private JClass clazzid = null;
	private long id = 0;

	public String genericsignature;
	public String signature;
	public String name;
	public boolean staticField = false;
	public Object staticValue = null;
	public int index = 0;
	public int modifiers = 0;

	public JField(JClass classid, long fid) {
		this.clazzid = classid;
		id = fid;

	}

	public long getFieldID() {
		return id;
	}

	@Override
	public Object get(JavaObject arg0) throws CorruptDataException,
			MemoryAccessException {
		
		if (staticField) {
			return staticValue;
		}
		
		CLogger.logr.log(Level.FINEST," " + this.name + " " + this.index + " "
				+ this.signature);
		
		if(arg0 == null) {
			throw new NullPointerException("Instance field passed null object.");
		}
		
		if(!(arg0 instanceof JObject)) {
			throw new IllegalArgumentException("Non CJVMTI JavaObject implementation was passed.");			
		}
		
		JObject job = (JObject) arg0;
	
		
		if(!((JObject)job).isInstance(clazzid)) {
			throw new IllegalArgumentException("Passed object incompatible with this fields' class.");
		}
			
		return job.getValue(id);
	}

	@Override
	public boolean getBoolean(JavaObject arg0) throws CorruptDataException,
			MemoryAccessException {
		if (staticValue != null) {
			return (Boolean) staticValue;
		}
		JObject job = (JObject) arg0;
		return job.getBooleanValue(id);
	}

	@Override
	public byte getByte(JavaObject arg0) throws CorruptDataException,
			MemoryAccessException {
		if (staticValue != null) {
			return (Byte) staticValue;
		}
		JObject job = (JObject) arg0;
		return job.getByteValue(id);
	}

	@Override
	public char getChar(JavaObject arg0) throws CorruptDataException,
			MemoryAccessException {
		if (staticValue != null) {
			return (Character) staticValue;
		}
		JObject job = (JObject) arg0;
		return job.getCharValue(id);
	}

	@Override
	public double getDouble(JavaObject arg0) throws CorruptDataException,
			MemoryAccessException {
		if (staticValue != null) {
			return (Double) staticValue;
		}
		JObject job = (JObject) arg0;
		return job.getDoubleValue(id);
	}

	@Override
	public float getFloat(JavaObject arg0) throws CorruptDataException,
			MemoryAccessException {
		if (staticValue != null) {
			return (Float) staticValue;
		}
		JObject job = (JObject) arg0;
		return job.getFloatValue(id);
	}

	@Override
	public int getInt(JavaObject arg0) throws CorruptDataException,
			MemoryAccessException {
		if (staticValue != null) {
			return (Integer) staticValue;
		}
		JObject job = (JObject) arg0;
		return job.getIntValue(id);
	}

	@Override
	public long getLong(JavaObject arg0) throws CorruptDataException,
			MemoryAccessException {
		if (staticValue != null) {
			return (Long) staticValue;
		}
		JObject job = (JObject) arg0;
		return job.getLongValue(id);
	}

	@Override
	public short getShort(JavaObject arg0) throws CorruptDataException,
			MemoryAccessException {
		if (staticValue != null) {
			return (Short) staticValue;
		}
		JObject job = (JObject) arg0;
		return job.getShortValue(id);
	}

	@Override
	public String getString(JavaObject arg0) throws CorruptDataException,
			MemoryAccessException {
		// Manually look for the string class
				
		
		if (!staticField && (arg0 == null) ) {
			throw new NullPointerException();
		}
		
		Object obj;
		if(staticField) {
			obj = staticValue;
		} else {
			obj = ((JObject)arg0).getValue(id);
		}

		// If the field value is null, return null.
		if (obj == null) {
			return null;
		}
		
		// Otherwise make sure we are looking at a JObject
		if(!(obj instanceof JObject)) {
			throw new IllegalArgumentException("Passed object must be CJVMTI JObject instance");
		}
		
		JObject stringObj = (JObject) obj;
		
		JavaClass stringClass = stringObj.getJavaClass();
		
		// java.lang.String is final.
		if(!stringClass.getName().equals("java/lang/String")) {
			throw new IllegalArgumentException("Field must be instance of java/lang/String");
		}
		

		JavaField valueField=null, countField=null, offsetField=null;
		for (JavaField field : stringObj.getJavaClass()
				.getDeclaredFields()) {
			// Assume this is the correct field
			String name = field.getName();
			if (name.equals("value")) {
				valueField = field;
			} else if (name.equals("count")) {
				countField = field;
			} else if (name.equals("offset")) {
				offsetField = field;
			}
			
			// Escape once we have the fields we need.
			if (valueField !=null && countField != null && offsetField != null) {
				break;
			}
		}
		
		if (valueField == null ||  countField == null || offsetField == null) {
			throw new CorruptDataException(new CorruptDataImpl("Unable to find relevant string fields"));
		}

		
		Character[] stringField = (Character[])((JObject)valueField.get(stringObj)).getObjArray();
		int offset = offsetField.getInt(stringObj);
		int count = countField.getInt(stringObj);
		char[] charArray = new char[count];
		for (int i = 0; i < count; i++){
			charArray[i] = stringField[offset+i];
		}
		
		String string = new String(charArray);
		
		CLogger.logr.log(Level.FINEST,"From field from "+this.getName()+" "+string);
		
		return string;
	}

	@Override
	public JavaClass getDeclaringClass() throws CorruptDataException,
			DataUnavailable {

		return clazzid;
	}

	@Override
	public int getModifiers() throws CorruptDataException {
		// TODO Auto-generated method stub
		return modifiers;
	}

	@Override
	public String getName() throws CorruptDataException {

		return name;
	}

	@Override
	public String getSignature() throws CorruptDataException {

		return signature;
	}

	@Override
	public void setValue(Object o) {
		staticValue = o;
	}

}