/*******************************************************************************
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/

package org.apache.kato.jvmti.javaruntime.model;

import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;

/**
 * Stores values in both a map and in a list.
 * 
 *
 * @param <K> Key type
 * @param <V> Value type
 */
public class ObjectMapList<K,V> {
	private HashMap<K,V> map=new HashMap<K,V>();
	private LinkedList<V> list=new LinkedList<V>();
	
	/**
	 * Returns unmodifiable List of the values.
	 * 
	 * @return List of map values.
	 */
	public List<V> values() {		
		return Collections.unmodifiableList(list);
	}
	
	/**
	 * Given a key, return the value associated with this.
	 * 
	 * @param key non-null object as a key.
	 * @return value or null if not found.
	 */
	public V get(K key) {
		if (key == null) {
			throw new NullPointerException("Key cannot be null.");
		}
		
		return map.get(key);
	}
	
	/**
	 * Associates a key with a value.
	 * 
	 * @param key  object to identify value. Must not be null.
	 * @param value  object to be associated with key. Must not be null.
	 * @return the previous value associated with the key, or null.
	 */
	public V put(K key, V value) {
		if (value == null) {
			throw new NullPointerException("Value must not be null.");
		}
		
		if (key == null) {
			throw new NullPointerException("Key must not be null.");
		}
		
		if(map.containsKey(key)) {
			V orig=map.get(key);
			
			// If the value is the same as what is already in the map and list
			// bypass removing the value from the list.
			// This takes far too long otherwise.
			if (orig.equals(value)) {
				return orig;
			}
			
			list.remove(orig);
		}
		
		list.add(value);
		return map.put(key, value);
	}
}
