/*******************************************************************************
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package org.apache.kato.jvmti.util;

import java.io.EOFException;
import java.io.File;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.util.LinkedHashMap;
import java.util.Map;

/*
 * This class could be modified to implement the write methods.
 * To do so the write methods could be modified to behave much like
 * the read methods. However, the cache would have to contain more
 * than just bytes such that they contain a "modified" boolean value
 * so the array can be written back to disk when they are pushed out
 * of the cache. Consideration will also need to be paid to length()
 * , setLength() and writing new cache blocks. 
 * 
 * CachedLinkHashMap.removeEldestEntry() is passed the entry to remove.
 * This could test for modification and then write it out if necessary.
 * 
 * The constructors must be modified to open a writable file, but
 * it's not necessary for it to be 
 * 
 * A flush method can be added to allow the cache to be written out
 * and the cache blocks to be marked as unmodified - class can
 * be modified to implement Flushable.
 * 
 * 
 * The close method will have to write out the remaining dirty
 * pages before closing.
 * 
 * 
 */

/**
 * <p>
 * A read-only RandomAccessFile with caching.
 * Implements an LRU cache to store blocks.
 * Blocks are implemented as byte arrays.
 * </p><p>
 * Only overrides the read methods that must be implemented.
 * The final read methods call the ones we have implemented here,
 * {@link readUTF()}, for instance.
 * </p><p>
 * {@link getFD()} and {@link getChannel()} won't benefit
 * from this class's caching and won't report the same
 * file position that would be returned by {@link getFilePointer()}.
 * Also, if the file pointer is modified by FileChannel, 
 * </p>
 * 
 */
final public class CachedRandomAccessFile extends RandomAccessFile {
	// The default block size, in bits. 4096 bytes.
	final static private int DEFAULT_BITS=12;
	final static private int DEFAULT_SIZE=16;// in MB
	private int cacheBits;  // Number of bits representing cache.
	private int blockSize; // block size.
	private long cacheUpperMask; // For extracting cache tag
	private long cacheLowerMask; // For extracting position in block
	private long length; // Length of the file
	private int numCacheEntries; // Number of cache entries to maintain
	private long position=0l; // Position in file.
	
	// The map from file locations to byte arrays.
	private CacheLinkedHashMap<Long,byte[]> cacheMap;
	
	/**
	 * Returns the block of bytes at a particular address. 
	 * The pos parameter will be within the block, not necessarily
	 * the beginning of the block.
	 * If necessary the block will be loaded from disk.
	 * 
	 * @param pos file pointer to return appropriate block for.
	 * @return byte[] containing blocks
	 * @throws IOException thrown if underlying read fails.
	 */
	byte[] getBlock(long pos) throws IOException {
		long tag = (pos & cacheUpperMask) >> cacheBits;
		
		byte[] block = cacheMap.get(tag);
		
		if (block == null) {
			pos &= cacheUpperMask;
			
			if (pos >= length) {
				throw new EOFException("Position "+pos+" is past end of file "+length);
			}
			
			if (blockSize > length) {
				// a file could be less than the blocksize in length.
				block = new byte[(int)length];				
			} else if (pos > (length-blockSize)) {
				// The last block in the file. File sizes are unlikely to be
				// a multiple of the cache block size.
				block = new byte[(int)(length-pos)];
			} else {
				block = new byte[blockSize];
			}
			
			super.seek(pos);
			super.read(block);
			cacheMap.put(tag, block);
		}
		
		return block;
	}
	
	/**
	 * Initialises the field variables and the cache.
	 * 
	 * @param cacheBits size of cache block in powers of 2
	 * @param cacheSize size of cache in megabytes
	 * @throws IOException if length() fails.
	 */
	private void init(int cacheBits, int cacheSize) throws IOException {
		this.cacheBits = cacheBits;
		this.blockSize = 1 << cacheBits;
		cacheUpperMask = ~(((long)blockSize) -1L);
		cacheLowerMask = this.blockSize-1;
		
		this.length = this.length();;
		
		numCacheEntries = (int)(cacheSize*1048576 / blockSize);
		
		this.cacheMap = new CacheLinkedHashMap<Long, byte[]>(numCacheEntries);
	}
	
	/**
	 * Creates a CachedRandomAccessFile open against the
	 * past File. 
	 * 
	 * @param file File to open.
	 * @throws IOException if file is not found.
	 */
	public CachedRandomAccessFile(File file)
	throws IOException {
		this(file, DEFAULT_BITS, DEFAULT_SIZE);
	}

	/**
	 * Creates a CachedRandomAccessFile open against the
	 * past file name.
	 * 
	 * @param name Name of file to open.
	 * @throws IOException if file is not found.
	 */
	public CachedRandomAccessFile(String name)
	throws IOException {
		this(name, DEFAULT_BITS, DEFAULT_SIZE);
	}

	/**
	 * <p>Creates a CachedRandomAccessFile open against the
	 * past File.
	 * </p><p>
	 * The cacheBits parameter is the number of bits that are
	 * used to address entries within cache blocks. The default
	 * value is 12, which equates to 4096 bytes per block. This
	 * is the amount of data read at once and stored within the cache.
	 * </p><p>
	 * The cacheSize is the maximum amount of memory occupied by the cache.
	 * This value is in megabytes. The default is 16, which is 16 megabytes.
	 * </p>
	 * 
	 * @param file File to open.
	 * @param cacheBits Number of bits representing cache block size
	 * @param cacheSize Cache size in megabytes.
	 * @throws IOException if file is not found.
	 */
	public CachedRandomAccessFile(File file, int cacheBits, int cacheSize)
	throws IOException {
		super(file, "r");
		
		init(cacheBits, cacheSize);
	}
	/**
	 * <p>Creates a CachedRandomAccessFile open against the
	 * past file name.
	 * </p><p>
	 * The cacheBits parameter is the number of bits that are
	 * used to address entries within cache blocks. The default
	 * value is 12, which equates to 4096 bytes per block. This
	 * is the amount of data read at once and stored within the cache.
	 * </p><p>
	 * The cacheSize is the maximum amount of memory occupied by the cache.
	 * This value is in megabytes. The default is 16, which is 16 megabytes.
	 * </p>
	 * 
	 * @param name Name fo file to open.
	 * @param cacheBits Number of bits representing cache block size
	 * @param cacheSize Cache size in megabytes.
	 * @throws IOException if file is not found.
	 */
	public CachedRandomAccessFile(String name, int cacheBits, int cacheSize)
	throws IOException {
		super(name, "r");
		
		init(cacheBits, cacheSize);
	}

	/**
	 * Reads a byte value. 
	 * 
	 * @return unsigned byte value, or -1 if beyond end of file.
	 * @throws IOException if underlying read fails.
	 */
	@Override
	public int read() throws IOException {
		// At end of file.
		if (position >= length) {
			return -1;
		}
		byte[] block = getBlock(position);
		long tmpPosition = position;
		position++;
		
		return ((int)block[(int)(tmpPosition & cacheLowerMask)]) & 0xff;		
	}
	
	@Override
	public int read(byte[] b) throws IOException {
		return read(b, 0, b.length);
	}
	
	@Override
	public int read(byte[] b, int off, int len) throws IOException {
		if (b == null) {
			throw new NullPointerException("byte[] b is null.");
		}

		if (len == 0) {
			return 0;
		}
		
		// We're starting beyond the end of the file.
		if (position >= length) {
			return -1;
		}
		
		// Check that the passed values are correct.
		if (off < 0) {
			throw new IndexOutOfBoundsException("off is < 0 ("+off+")");
		} else if (len < 0) {
			throw new IndexOutOfBoundsException("len  is < 0 ("+len+")");
		} else if (len > (b.length-off)) {
			throw new IndexOutOfBoundsException("len("+len+") is > b.length("+b.length+")-off("+off+")");
		}
		
		
		byte[] block = getBlock(position);

		int blockPos = (int) (position & cacheLowerMask);

		// Simple case - one arraycopy suffices.
		// block.length is used
		if( (blockPos+len) < block.length) {
			System.arraycopy(block, blockPos, b, off , len);
			position += len;
			return len;
		} else {
			int remaining = len;
			int arrayPos = 0;
			int toRead;

			
			while (remaining > 0) {
				// Read either a block, or the bytes remaining.
				if ( (blockPos+remaining) >= block.length) {
					toRead = (int) block.length - blockPos;
				} else {
					toRead = (int) remaining;
				}

				System.arraycopy(block, blockPos, b, off+arrayPos , toRead);

				position = position + toRead;
				arrayPos += toRead;

				remaining -= toRead;

				// check if we've reached the end of the file.
				if (position >= length) {					
					break;
				} else if (remaining > 0) {
					block = getBlock(position);				
					blockPos = 0; // The next block will always be on a
				}				
			}
			return arrayPos;
		}
	}
	
	
	/**
	 * Returns the position within the file.
	 * The underlying RandomAccessFile file pointer may be different
	 * from this value as seeks are carried out only when cache blocks
	 * are read from disk.
	 * 
	 */
	@Override
	public long getFilePointer() throws IOException {
		return position;
	}
	
	/**
	 * Repositions the file-pointer offset to where the
	 * next read will occur. This does not affect the underlying
	 * RandomAccessFile unless a cache block has to be read in.
	 * 
	 * 
	 * @throws IOException is throws if pos is < 0. 
	 */
	@Override
	public void seek(long pos) throws IOException {
		if (pos < 0l) {
			throw new IOException("Position "+pos+" is < 0");
		}
		this.position = pos;
	}

	/**
	 * Skips over some bytes without reading them.
	 * This doesn't affect the RandomAccessFile
	 * position. 
	 * 
	 * @param n number of bytes to skip over.
	 * @returns number of bytes skipped
	 */
	@Override
	public int skipBytes(int n) throws IOException {
		// Place position at end of the file.
		if (n <= 0) {
			return 0;
		} else if (position + n >= length) {
			long tmpPosition = position;
			position = length;
			
			return (int) (length - (long)tmpPosition);
		} 

		position += n;
		
		return n;
	}
	
	/**
	 * Closes the open file and empties the cache. 
	 * As the cache is empty after this, future accesses
	 * will try to load the cache blocks from disk, and so
	 * will fail. Operations affecting the file position will
	 * continue to work without exceptions being thrown.
	 * 
	 * @throws IOException if underlying close throws IOException.
	 */
	@Override
	public void close() throws IOException {
		super.close();
		
		// Throw away cache with an empty alternative.
		// Future attempts to populate the cache will result in IOExceptions.
		this.cacheMap = new CacheLinkedHashMap<Long, byte[]>(0);
	}
	
	// Don't allow any write operations
	
	/**
	 * No write operations are supported, so an IOException is always
	 * thrown.
	 * 
	 * @throws IOException is thrown unconditionally. 
	 */
	@Override
	public void write(int b) throws IOException {
		throw new IOException("CachedRandomAccessFile is not writable");
	}

	/**
	 * No write operations are supported, so an IOException is always
	 * thrown.
	 * 
	 * @throws IOException is thrown unconditionally. 
	 */
	@Override
	public void write(byte[] b) throws IOException {
		throw new IOException("CachedRandomAccessFile is not writable");
	}

	/**
	 * No write operations are supported, so an IOException is always
	 * thrown.
	 * 
	 * @throws IOException is thrown unconditionally.
	 */
	@Override
	public void write(byte[] b, int off, int len) throws IOException {
		throw new IOException("CachedRandomAccessFile is not writable");
	}	
	
	/**
	 * No write operations are supported, so an IOException is always
	 * thrown.
	 * 
	 * @throws IOException is thrown unconditionally.
	 */
	@Override
	public void setLength(long newLength) throws IOException {
		throw new IOException("CachedRandomAccessFile is not writable");
	}

	/**
	 * LinkedHashMap for our caching purposes.
	 * Used as a LRU cache, holding up to a fixed
	 * number of cache entries.  
	 *	
	 * @param <K> Key type
	 * @param <V> Value type
	 */
	@SuppressWarnings("serial")
	private final static class CacheLinkedHashMap<K,V> extends LinkedHashMap<K,V> {
		private int maxCacheEntries;
		
		/**
		 * Creates a LinkedHashMap with access ordering.
		 * The oldest entry is the oldest entry to be put or
		 * got.
		 * Takes the number of entries to cache.
		 * 
		 * 
		 * @param maxCacheEntries maximum number of entries to hold.
		 */
		public CacheLinkedHashMap(int maxCacheEntries) {
			super(maxCacheEntries, 0.75f, true);
			this.maxCacheEntries = maxCacheEntries;
		}
		
		
		/**
		 * Overrides the default implementation to contain
		 * the cache size. 
		 * 
		 * @return true if cache entry is to be removed.
		 * @param entry oldest entry to handle.
		 */
		@Override
		protected boolean removeEldestEntry(Map.Entry<K,V> entry) {
			return size() > maxCacheEntries;
		}
	}
}
