/*******************************************************************************
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package org.apache.kato.hprof.java;


import java.util.LinkedList;
import java.util.List;

import javax.tools.diagnostics.image.CorruptDataException;
import javax.tools.diagnostics.image.DataUnavailable;
import javax.tools.diagnostics.image.ImagePointer;
import javax.tools.diagnostics.image.MemoryAccessException;
import javax.tools.diagnostics.runtime.java.JavaClass;
import javax.tools.diagnostics.runtime.java.JavaHeap;
import javax.tools.diagnostics.runtime.java.JavaObject;

import org.apache.kato.common.BasicType;
import org.apache.kato.hprof.datalayer.IGCPrimitiveArrayHeapDumpRecord;
import org.apache.kato.hprof.image.CorruptDataImpl;

public class JavaPrimitiveArrayImpl implements JavaObject {
	private IJavaHeapInternal heap;
	private IGCPrimitiveArrayHeapDumpRecord record;

	public JavaPrimitiveArrayImpl(IJavaHeapInternal heap, IGCPrimitiveArrayHeapDumpRecord record) {
		this.heap = heap;
		this.record = record;
	}
	@Override
	public void arraycopy(int srcStart, Object dst, int dstStart, int length)
			throws CorruptDataException, MemoryAccessException {
		if (srcStart < 0 || srcStart > getArraySize()) {
			throw new IndexOutOfBoundsException("srcStart must be between 0 and " + getArraySize()+", not " + srcStart);
		}
		
		if (srcStart+length > getArraySize()) {
			throw new IndexOutOfBoundsException("srcStart+length must be <="+getArraySize()+", srcStart=" + srcStart+", length="+length);			
		}
		if (dstStart < 0) {
			throw new IndexOutOfBoundsException("dstStart must be >=0, not " + dstStart);
		}

		if (dst == null) {
			throw new NullPointerException("Destination array must not be null");
		}
		if (!dst.getClass().isArray()) {
			throw new IllegalArgumentException("Destination array must be an array, not "+dst.getClass().getName());			
		}
		
		if (length < 0) {
			throw new IndexOutOfBoundsException("length must be >=0, not " + length);
		}
		
		switch(record.getElementType()) {
		case BasicType.BOOLEAN:
		{
			if(dst.getClass().getComponentType().getName() != "boolean") {
				throw new IllegalArgumentException("Destination array wrong type - must be boolean");
			}
			boolean[] array = (boolean[]) dst;
			checkArrayBounds(dstStart, array.length, length);
			
			for (int i=0; i < length; i++) {
				array[dstStart+i] = record.getBooleanElement(srcStart+i);
			}
			break;
		}
		case BasicType.BYTE:
		{
			if(dst.getClass().getComponentType().getName() != "byte") {
				throw new IllegalArgumentException("Destination array wrong type - must be byte");
			}
			byte[] array = (byte[]) dst;
			checkArrayBounds(dstStart, array.length, length);
			
			for (int i=0; i < length; i++) {
				array[dstStart+i] = (byte) record.getByteElement(srcStart+i);
			}
			break;
		}
		case BasicType.SHORT:
		{
			if(dst.getClass().getComponentType().getName() != "short") {
				throw new IllegalArgumentException("Destination array wrong type - must be short");
			}
			short[] array = (short[]) dst;
			checkArrayBounds(dstStart, array.length, length);
			
			for (int i=0; i < length; i++) {
				array[dstStart+i] =  record.getShortElement(srcStart+i);
			}
			break;
		}
		case BasicType.CHAR:
		{
			if(dst.getClass().getComponentType().getName() != "char") {
				throw new IllegalArgumentException("Destination array wrong type - must be char");
			}
			char[] array = (char[]) dst;
			checkArrayBounds(dstStart, array.length, length);
			
			for (int i=0; i < length; i++) {
				array[dstStart+i] =  record.getCharElement(srcStart+i);
			}
			break;
		}
		case BasicType.INT:
		{
			if(dst.getClass().getComponentType().getName() != "int") {
				throw new IllegalArgumentException("Destination array wrong type - must be int");
			}
			int[] array = (int[]) dst;
			checkArrayBounds(dstStart, array.length, length);
			
			for (int i=0; i < length; i++) {
				array[dstStart+i] =  record.getIntElement(srcStart+i);
			}
			break;
		}
		case BasicType.LONG:
		{
			if(dst.getClass().getComponentType().getName() != "long") {
				throw new IllegalArgumentException("Destination array wrong type - must be long");
			}
			long[] array = (long[]) dst;
			checkArrayBounds(dstStart, array.length, length);
			
			for (int i=0; i < length; i++) {
				array[dstStart+i] =  record.getLongElement(srcStart+i);
			}
			break;
		}
		case BasicType.FLOAT:
		{
			if(dst.getClass().getComponentType().getName() != "float") {
				throw new IllegalArgumentException("Destination array wrong type - must be float");
			}
			float[] array = (float[]) dst;
			checkArrayBounds(dstStart, array.length, length);
			
			for (int i=0; i < length; i++) {
				array[dstStart+i] =  record.getFloatElement(srcStart+i);
			}
			break;
		}
		case BasicType.DOUBLE:
		{
			if(dst.getClass().getComponentType().getName() != "double") {
				throw new IllegalArgumentException("Destination array wrong type - must be double");
			}
			double[] array = (double[]) dst;
			checkArrayBounds(dstStart, array.length, length);
			
			for (int i=0; i < length; i++) {
				array[dstStart+i] =  record.getDoubleElement(srcStart+i);
			}
			break;
		}
		default:
			throw new CorruptDataException(new CorruptDataImpl("Primitive array has invalid type of "+record.getElementType()));
		}
	}

	private static void checkArrayBounds(int arrayStart, int arrayLength, int dataLength) {
		if ( arrayStart + dataLength > arrayLength) {
			throw new IndexOutOfBoundsException("Destination array of length "+ arrayLength+" can't" 
					+ " accomodate array copy with start of "+arrayStart+" and length of "+ dataLength);
		}
	}
	@Override
	public int getArraySize() throws CorruptDataException {
		return record.getNumberOfElements();
	}

	@Override
	public long getHashcode() throws DataUnavailable, CorruptDataException {
		// TODO Auto-generated method stub
		return 0;
	}

	@Override
	public JavaHeap getHeap() throws CorruptDataException, DataUnavailable {
		return heap;
	}

	@Override
	public ImagePointer getID() {
		return heap.getImagePointer(record.getID());
	}

	@Override
	public JavaClass getJavaClass() throws CorruptDataException {
		return heap.getPrimitiveArrayClass(record.getElementType());
	}

	@Override
	public long getPersistentHashcode() throws DataUnavailable,
			CorruptDataException {
		throw new DataUnavailable("Persistent hashcode unknown.");	}

	@Override
	public List getReferences() {
		return new LinkedList();
	}

	@Override
	public List getSections() {
		return new LinkedList();
	}

	@Override
	public long getSize() throws CorruptDataException {
		return 0;
	}

	@Override
	public boolean isArray() throws CorruptDataException {
		return true;
	}

	/**
	 * Compares identity. This object may be created multiple times.
	 * Uses object ID for comparing identity. 
	 * We don't check runtime identity.
	 */
	@Override
	public boolean equals(Object obj) {
		if (obj == null) {
			return false;
		}
		
		if (!(obj instanceof JavaPrimitiveArrayImpl)) {
			JavaPrimitiveArrayImpl other = (JavaPrimitiveArrayImpl) obj;
			
			if (other.getObjectID() == record.getID()) {
				return true;
			}
		}
		
		return false;
	}
	
	/**
	 * Return object ID of record on disk.
	 * @return
	 */
	long getObjectID() {
		return record.getID();
	}
}
