/*******************************************************************************
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package org.apache.kato.hprof.java;

import java.util.LinkedList;
import java.util.List;

import javax.tools.diagnostics.image.CorruptDataException;
import javax.tools.diagnostics.image.DataUnavailable;
import javax.tools.diagnostics.image.ImagePointer;
import javax.tools.diagnostics.image.DiagnosticException;
import javax.tools.diagnostics.runtime.java.JavaClass;
import javax.tools.diagnostics.runtime.java.JavaLocation;
import javax.tools.diagnostics.runtime.java.JavaMethod;
import javax.tools.diagnostics.runtime.java.JavaStackFrame;
import javax.tools.diagnostics.runtime.java.JavaVariable;

import org.apache.kato.hprof.IJavaStackFrame;
import org.apache.kato.hprof.image.ImagePointerImpl;

/**
 * 	A stack frame, wraps a {@link IJavaHeapInternal}.
 *
 */
public class JavaStackFrameImpl implements JavaStackFrame {
	private IJavaStackFrame frame;
	private IJavaHeapInternal heap;
	
	public JavaStackFrameImpl(IJavaHeapInternal heap, IJavaStackFrame frame) {
		this.frame = frame;
		this.heap = heap;
	}
	
	/**
	 * There is no stored frame bas pointer.
	 * Returns frame ID instead.
	 * 
	 * FIXME Would like a DataUnavailable exception here.
	 */
	@Override
	public ImagePointer getBasePointer() throws CorruptDataException {		
		return heap.getImagePointer(frame.getID());
	}

	@Override
	public List getHeapRoots() {
		// TODO Need implementation here.
		return null;
	}

	private JavaLocation location = new JavaLocationImpl();
	@Override
	public JavaLocation getLocation() throws CorruptDataException {
		return location;
	}

	/**
	 * Implements a JavaLocation
	 * There is a 1:1 relationship with JavaStackFrames. 
	 *
	 */
	private class JavaLocationImpl implements JavaLocation {
		
		/**
		 * For testing equality.
		 * 
		 * @return IJavaStackFrame of this location
		 */
		IJavaStackFrame getFrame() {
			return frame;			
		}
		
		
		@Override
		public ImagePointer getAddress() throws CorruptDataException {
			return heap.getImagePointer(frame.getID());
		}

		
		/**
		 * Returns 1 if the method is compiled, otherwise returns 0
		 */
		@Override
		public int getCompilationLevel() throws CorruptDataException {
			int level = frame.getLineNumber();
		
			return level == -2 ? 1 : 0;			
		}

		@Override
		public String getFilename() throws DataUnavailable,
				CorruptDataException {			
			return frame.getSourceFileName();
		}

		@Override
		public int getLineNumber() throws DataUnavailable, CorruptDataException {
			int level = frame.getLineNumber();
			
			if (level < 0) {
				throw new DataUnavailable("No line number information.");
			}
			return level;
		}

		private JavaMethodImpl method = new JavaMethodImpl();
		
		@Override
		public JavaMethod getMethod() throws CorruptDataException {
			return method;
		}
		
		public boolean equals(Object obj) {
			if (obj == null) return false;
			
			if (! (obj instanceof JavaLocationImpl)) {
				return false;
			}
			
			JavaLocationImpl otherLocation = (JavaLocationImpl) obj;
			
			return otherLocation.getFrame().equals(frame);
		}
		
		// TODO
		public String toString() {
			return "a string";
		}
		
	}

	/**
	 *  A dummy method. 
	 *
	 */
	private class JavaMethodImpl implements JavaMethod {

		/**
		 * HProf has no bytecode sections.
		 * 
		 * @return empty Iterator
		 */
		@Override
		public List getBytecodeSections() {			
			return new LinkedList();
		}

		/**
		 * HProf has no compiled code sections.
		 * @return empty Iterator
		 */
		@Override
		public List getCompiledSections() {
			return new LinkedList();
		}

		@Override
		public JavaClass getDeclaringClass() throws CorruptDataException,
				DataUnavailable {			
			return heap.getJavaClassByID(frame.getMethodClass().getClassObjectID());
		}

		/**
		 * HProf doesn't store modifiers. 
		 * A method will never be reported as static.
		 * 
		 * @return 0
		 */
		@Override		
		public int getModifiers() throws CorruptDataException {
			return 0;
		}

		@Override
		public String getName() throws CorruptDataException {
			return frame.getMethodName();
		}

		@Override
		public String getSignature() throws CorruptDataException {
			return frame.getMethodSignature();
		}
		
		public boolean equals(Object obj) {
			if (obj == null) return false;
			
			if (! (obj instanceof JavaMethodImpl)) {
				return false;
			}
			
			JavaMethodImpl otherMethod = (JavaMethodImpl) obj;
			
			try {
				if (otherMethod.getDeclaringClass().equals(getDeclaringClass()) &&
						otherMethod.getName().equals(getName()) &&
						otherMethod.getSignature().equals(getSignature())) {
					return true;
				}
			} catch (CorruptDataException e) {
			} catch (DataUnavailable e) {
			}
			
			return false;
		}
		
		/**
		 * @return String in format "classname.methodname(signature)";
		 */
		public String toString() {
			try {
				return getDeclaringClass().getName()+"."+getName()+getSignature();
			} catch (DiagnosticException e) {				
				return "Exception on JavaMethodImpl.toString()+"+ e.getMessage();
			}  
		}

		@Override
		public List<JavaVariable> getVariables()  {
			return new LinkedList<JavaVariable>();
		}
	}

	@Override
	public Object getVariable(int slot) throws CorruptDataException,
			DataUnavailable, IndexOutOfBoundsException {
		throw new DataUnavailable();
	}

	@Override
	public List getVariables() {
		
		return new LinkedList();
	}
	
}
