/*******************************************************************************
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package org.apache.kato.tools.jdi;
import java.io.ByteArrayOutputStream;
import java.io.PrintStream;
import java.util.Vector;

/* JDWPServer
 * 
 * JDWPServer provides the high-level core to JDI.  It starts the socket server, and provides
 * facilities to convert bytes to and from the Packet class, as well as calling the outbound socket.
 * The sockets directly call writeRawPacket(...) which converts bytes into a Packet Vector, which
 * is one by one fired at the cache.  The cache has access to the KatoReader, which converts the
 * Packet into a set of Kato calls (basically a JDWP <--> Kato layer), and then assembles a new
 * packet, and returns back up to writeRawPacket(...) which converts the Packet back to bytes and
 * explicitly calls the write on the socket and back to the client
 * 
 */

public class JDWPServer extends CacheReader{
	private int port;
	private JDISocketServer sockServe;
	//Keep track of the internal errors we've had
	private int lastSeq;
	private JDILogger logr;
	private boolean lastClientTerm = false;
	private byte []leftOverBytes;
	
	public JDWPServer(JDILogger logger, int port) throws Exception{
		logr = logger;
		this.port = port;
		logr.log(JDILogger.LEVEL_VERYVERBOSE, "JDWPServer 1.0"); //$NON-NLS-1$


	}

	public JDISocketServer getSocketServer(){
		return sockServe;
	}

	public void setTerminateLastClient(boolean setting){
		if (sockServe == null){
			lastClientTerm = setting;
		}else{
			sockServe.setTerminateLastClient(setting);
		}
	}

	public boolean getTerminateLastClient(){
		if (sockServe == null)
			return lastClientTerm;
		return sockServe.getTerminateLastClient(); 


	}

	public void start() throws Exception{
		logr.log(JDILogger.LEVEL_VERYVERBOSE, "JDWPServer Start"); //$NON-NLS-1$

		try{
			//Create the new server stuff
			sockServe = new JDISocketServer(this, port);

			//Start the server
			sockServe.start();

			//Set the preferences set earlier for disconnection policy
			sockServe.setTerminateLastClient(lastClientTerm);
		}catch(Exception exxy){
			//If we error out, shut down Kato
			getCache().getProvider().stop();

			//And log it
			logr.logError(JDILogger.LEVEL_NORMAL, "Could not start server"); //$NON-NLS-1$
			exxy.printStackTrace(new PrintStream(logr.getErrorStream()));
			throw exxy;
		}

	}

	public int getLastSequence(){
		return lastSeq;
	}

	//This is takes the raw bytes incoming from the server
	//And figures out what to do with them
	public void writeRawPacket(byte []inBytes, JDISocket inVSock){
		//Handshake?  Let's check
		if (inBytes.length == 14 && bytesEqual("JDWP-Handshake", inBytes)){ //$NON-NLS-1$
			inVSock.write(inBytes);
		}else{
			//Here we take a look at the bytes to see if they make sense
			//And to see if we can actually make a coherent command out of them
			Vector<CommandPacket> vctr = generateCommandPacket(inBytes);
			while(vctr.size() > 0){
				CommandPacket cmd = vctr.firstElement();
				lastSeq = cmd.getSequence();
				cmd.setVSock(inVSock);
				vctr.remove(0);
				//most commands have a reply
				long time = System.currentTimeMillis();
				ReplyPacket rply = askPacket(cmd);
				time = System.currentTimeMillis() - time;
				if (time > 250){
					logr.log(JDILogger.LEVEL_VERBOSE, "Possible performance problem! (" + time + "ms, seq " + cmd.getSequence() + ")"); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
				}
				if (rply != null){
					//Set the resultant upstream socket
					//If they haven't set one
					if (rply.getVSock() == null){
						rply.setVSock(inVSock);
					}
					sendReplyPacket(rply);
					rply.setVSock(null);
				}else{
					logr.log(JDILogger.LEVEL_VERBOSE, "Unmatched: seq=" + cmd.getSequence()); //$NON-NLS-1$
				}

			}
		}
	}

	public void stop(){
		logr.log(JDILogger.LEVEL_VERBOSE, "Shutting down server"); //$NON-NLS-1$
		//We forcefully close the listening server socket
		//And set a flag that quietly shuts stuff down
		//Which triggers a join in this.start()
		if (sockServe != null){
			sockServe.unlisten();
		}
		closeCache();
		sockServe = null;
	}

	//This sends back a reply packet
	public void sendReplyPacket(ReplyPacket rpckt){
		if (rpckt.getVSock() == null){
			logr.logError(JDILogger.LEVEL_VERBOSE, "Socket seq " + rpckt.getSequence() + " has no known destination"); //$NON-NLS-1$ //$NON-NLS-2$
			stop();
			return;
		}
		byte []foo = rpckt.toBytes();
		rpckt.getVSock().write(foo);

	}

	public void disconnect(JDISocket ivsock){
		//Disconnect ivsock
		ivsock.disconnect();
	}

	//This sends back a command packet
	public void sendCommandPacket(CommandPacket cpckt){

		cpckt.getVSock().write(cpckt.toBytes());


	}

	//We originally returned a packet, but sometimes some of the packets got mashed together.
	//I now (try) to decode them all, and put them in a vector
	//Once in the vector, they're returned back for processing, starting with the first received.
	//Once processing is done, I return back here to get the next bunch.
	private Vector<CommandPacket> generateCommandPacket(byte []inBytes){
		Vector<CommandPacket> outData = new Vector<CommandPacket>();
		if (leftOverBytes != null && leftOverBytes.length > 0){
			logr.logError(JDILogger.LEVEL_VERYVERBOSE, "Attempting to glue last request to current"); //$NON-NLS-1$
			try{
				ByteArrayOutputStream bos = new ByteArrayOutputStream();
				bos.write(leftOverBytes);
				bos.write(inBytes);
				inBytes = bos.toByteArray();
			}catch(Exception exxy){}
			leftOverBytes = null;
		}
		int dataLeft = inBytes.length;
		while(dataLeft > 0){
			//Minimum packet size
			if (dataLeft < 11){
				logr.logError(JDILogger.LEVEL_VERYVERBOSE, "Truncated packet does not contain header? ("+dataLeft+")"); //$NON-NLS-1$ //$NON-NLS-2$
				try{
					ByteArrayOutputStream bos = new ByteArrayOutputStream();

					bos.write(inBytes);

					leftOverBytes = bos.toByteArray();
				}catch(Exception exxy){
					leftOverBytes = null;
				}
				return outData;
			}

			int length = KatoReader.createIntFromBytes(inBytes, 0, 4);
			if (length > inBytes.length){
				logr.logError(JDILogger.LEVEL_VERYVERBOSE, "Truncated packet does not have correct size? (got "+inBytes.length+", size "+length+")"); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
				try{
					ByteArrayOutputStream bos = new ByteArrayOutputStream();

					bos.write(inBytes);

					leftOverBytes = bos.toByteArray();
				}catch(Exception exxy){
					leftOverBytes = null;
				}
				return outData;
			}
			int sequence = KatoReader.createIntFromBytes(inBytes, 4, 4);
			int flags = inBytes[8];
			if (flags != 0){
				logr.log(JDILogger.LEVEL_VERBOSE, "ReplyPacket received"); //$NON-NLS-1$
			}else{
				int cmdset = inBytes[9];
				int cmd = inBytes[10];
				CommandPacket cmdpckt = new CommandPacket(sequence, flags, cmdset, cmd);
				byte []dataSegment = getBytesFromRange(inBytes, 11, length);
				cmdpckt.setData(dataSegment);
				outData.add(cmdpckt);
			}
			try{
				ByteArrayOutputStream bos = new ByteArrayOutputStream();

				bos.write(inBytes, length, dataLeft - length);

				inBytes = bos.toByteArray();
			}catch(Exception exxy){
			}
			dataLeft = inBytes.length;


		}
		return outData;
	}

	private byte[] getBytesFromRange(byte []inBytes, int lower, int higher){
		byte []outBytes = new byte[higher - lower];
		for(int i = lower; i < higher; i++){
			outBytes[i - lower] = inBytes[i]; 
		}
		return outBytes;
	}

	private boolean bytesEqual(String comp, byte []inBytes){
		if (comp.length() != inBytes.length)
			return false;
		char []str = new char[comp.length()];
		str = comp.toCharArray();
		for (int i = 0; i < comp.length(); i++){
			if (str[i] != inBytes[i])
				return false;
		}
		return true;
	}

	public ReplyPacket askPacket(CommandPacket inpckt){
		logr.log(JDILogger.LEVEL_VERYVERBOSE, "JDWPServer.askPacket(" + inpckt.getSequence() + "," + inpckt.getFlags() + "," + inpckt.getCommandSet() + "," + inpckt.getCommand() + ")"); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$ //$NON-NLS-5$
		JDICacher localCache = getCache();
		return localCache.query(inpckt);
	}


}
