/*******************************************************************************
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/

package org.apache.kato.tck.harness.scenario;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.PrintStream;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.Iterator;
import java.util.List;

import org.apache.kato.tck.api.TCKException;
import org.apache.kato.tck.harness.MapCollector;
import org.apache.kato.tck.harness.ValueCollector;

/**
 * 
 */
public class ConfigurationExtractor implements IScenarioSetupHandler{
	private static File home=new File(System.getProperty("user.home"));
	private MapCollector envvars = null;
	private ValueCollector vmargs = null;
	private String configDirName=null;
	PrintStream  envout=null;
	PrintStream  jvmout=null;
	
	public ConfigurationExtractor() {
		envvars=new MapCollector();
		vmargs=new ValueCollector();
	
	}
	private void extractEnvVarConfig(ScenarioElement element, Method method) {
		try {
			Class[] parms = method.getParameterTypes();
			if (parms == null || parms.length == 0) {
				Class returnType = method.getReturnType();
				if (returnType.equals(String[].class)) {
					String[] result = (String[]) method.invoke(element, null);
					if(result!=null) {
						for(int i=0;i<result.length;i++) {
							envvars.collect(result[i], method);
						}
					}
	
				}
			}
	
		} catch (Exception e) {
			e.printStackTrace();
			throw new TCKException("Exception using envvar method " + method, e);
	
		}
	}

	/**
	 * Extract configuration information for an instance of IScenarioElement
	 * All declared methods are scanned and those that are
	 * prefixed with "env"  or "jvmarg" and have a String[] return value ,are public
	 * and not static are considered. If the method takes parameters it is ignored
	 * 
	 *
	 */
	public void execute(ScenarioElement element) {
		// get methods
		Class me = element.getClass();
		Method m[] = me.getDeclaredMethods();
	
		// run all methods where the method name starts with configure
		for (int j = 0; j < m.length; j++) {
			Method method = m[j];
			
			// ignore non public and methods that return values
			if(Modifier.isPublic(method.getModifiers()) ==false) continue;
			if(Modifier.isStatic(method.getModifiers()) ==true ) continue;
			if(method.getReturnType().equals(String[].class)==false) continue;
			if(method.getParameterTypes().length>0) continue;
			
			String methodName = m[j].getName();
			if (methodName.startsWith("env")) {
	
				extractEnvVarConfig(element, method);
				
			} else if (methodName.startsWith("jvmarg")) {
				reportVMVar(element, method);
			}
	
		}
	}

	private void reportVMVar(ScenarioElement element, Method method) {
		try {
			Class[] parms = method.getParameterTypes();
			if (parms == null || parms.length == 0) {
				Class returnType = method.getReturnType();
				if (returnType.equals(String[].class)) {
					String[] result = (String[]) method.invoke(element, null);
					if(result!=null) {
						for(int i=0;i<result.length;i++) {
							vmargs.collect(result[i], method);
						}
					}
				}
			}
	
		} catch (Exception e) {
			e.printStackTrace();
			throw new TCKException("Exception using jvmarg method " + method, e);
	
		}
	}

	private void writeList(List entries, PrintStream out) {
		
		Iterator i=entries.iterator();
		while(i.hasNext()) {
			out.println(i.next());
		}
		
	}
	/**
	 * @param configdir
	 */
	public void setConfigDirectory(String configdir) {
		this.configDirName=configdir;
		
			
			
			
		
		
	}
	
	/**
	 * @param envfile
	 * @return
	 * @throws FileNotFoundException 
	 */
	private PrintStream buildStream(File out) throws FileNotFoundException {
		
		FileOutputStream fos=new FileOutputStream(out);
		return new PrintStream(fos);
		
	}
	/* (non-Javadoc)
	 * @see org.apache.kato.tck.harness.scenario.IScenerioSetupHandler#close()
	 */
	public void close() {
		writeList(envvars.asList(),envout);
		writeList(vmargs.asList(),jvmout);
		envout.close();
		jvmout.close();
		
		
	}
	/* (non-Javadoc)
	 * @see org.apache.kato.tck.harness.scenario.IScenerioSetupHandler#open()
	 */
	public void open() throws IOException {
		if(configDirName!=null) {
			home=new File(configDirName);
			home.mkdirs();
		}
		
		File envfile=new File(home,"tck.envfile.txt");
		File vmfile=new File(home,"tck.jvmfile.txt");
		
		envout=buildStream(envfile);
		jvmout=buildStream(vmfile);
		
	
	
		
	}
	/* (non-Javadoc)
	 * @see org.apache.kato.tck.harness.scenario.IScenerioSetupHandler#getStatus()
	 */
	public String getStatus() {
		return "Scenerio Configuration : environmental variables="+envvars.size()+" jvm arguments="+vmargs.size();
	}

}
