/*******************************************************************************
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/

package org.apache.kato.tck.harness.scenario;

import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.CommandLineParser;
import org.apache.commons.cli.GnuParser;
import org.apache.commons.cli.HelpFormatter;
import org.apache.commons.cli.Option;
import org.apache.commons.cli.OptionBuilder;
import org.apache.commons.cli.Options;
import org.apache.kato.tck.api.TCKException;

/**
 * 
 */
public class ScenarioLauncher {
	private static final String DUMPTYPE = "dumptype";

	/**
	 * 
	 */
	private static final String CLASS = "class";
	
	/**
	 * 
	 */
	private static final String MODE2 = "mode";

	/**
	 * 
	 */
	private static final String CONFIGDIR = "configdir";

	/**
	 * 
	 */
	private static final String DUMPTRIGGER = "dumptrigger";

	
	
	private IScenarioSetupHandler handler=null;
	private List classestoload=null;
	
	private ScenarioLauncher() {
		
	}
	/**
	 * Builds a Command line configuration object for use in parsing
	 * the command line
	 * @return
	 */
	private static Options buildOptions() {
		Options options = new Options();
		Option scenario = OptionBuilder.withArgName("id").hasArg()
				.withDescription("scenario to run").create("scenario");
		options.addOption(scenario);

		Option clazz = OptionBuilder.withArgName("classname").hasArg()
				.withDescription("classname to run").create(CLASS);
		options.addOption(clazz);

		Option mode = OptionBuilder.withArgName("modetype").hasArg()
				.withDescription("run mode: config or run").create(MODE2);
		mode.setRequired(true);
		options.addOption(mode);
		
		Option home = OptionBuilder.withArgName("file").hasArg()
		.withDescription("directory to store configuration (defaults to '.tck' in user home)").create(CONFIGDIR);
		options.addOption(home);
		
		Option dumpclass=OptionBuilder.withArgName("dumpclassname").hasArg()
		.withDescription("Dump trigger class name (defaults to jvm level specific dump trigger)").create(DUMPTRIGGER);
		options.addOption(dumpclass);
		
		Option dumptype = OptionBuilder.withArgName(DUMPTYPE).hasArg()
		.withDescription("dump type to generate").create(DUMPTYPE);
		options.addOption(dumptype);

		return options;
		
	}
	
	/**
	 * Load a class by name
	 * 
	 * @param value dot formed fully qualified name of class to load
	 * @return class loader or null if not available
	 */
	private static Class loadScenarioClass(String value) {

		try {
			Class clazz = Class.forName(value);

			return clazz;
		} catch (ClassNotFoundException e) {
			return null;
		}

	}
	/**
	 * @param args
	 */
	public static void main(String[] args) {
		
		ScenarioLauncher l = new ScenarioLauncher();
		try {
		IScenarioSetupHandler handler=l.createHandler(args);
		if(handler==null) {
			System.exit(-1);
		}
		
		ScenarioAnchor.setHandler(handler);
		
		handler.open();
		
		// instantiate objects which will call back 
		// to handler to configure or setup
		
		l.executeScenarioSetup();
		
		System.out.println(handler.getStatus());
		
		handler.close();
		
		
		
		
		} catch(Throwable t) {
			t.printStackTrace();
			l.printHelp();
			
			System.exit(-2);
		}
		
		
		System.exit(0);

	}
	

	

	


	private Options options = null;

	

	
	/**
	 * Reports error and displays launcher instructions
	 * 
	 * @param msg message to report 
	 */
	
	private void error(String msg) {
		System.err.println("Setup failed : " + msg);
		printHelp();

	}

	/**
	 * run configuration proper by instantiating the classes. Classes are at
	 * liberty to call back into the scenario launcher for purposes of using the
	 * thread element configurator
	 **/

	private boolean executeScenarioSetup() {

		

		Iterator i = classestoload.iterator();
		while (i.hasNext()) {
			Class clazz = (Class) i.next();
			
			try {
				clazz.newInstance();
				
				
			} catch (Throwable e) {
				e.printStackTrace();
				error("unable to construct scenario for " + clazz);
				return false;
			}

		}

		return true;
	}

	
	
	/**
	 * 
	 */
	private void printHelp() {

		HelpFormatter formatter = new HelpFormatter();
		formatter.printHelp(getClass().getName(), options);
		
	}

	/**
	 * Configure the system based on the scenarios declared in the input
	 * parameters
	 * 
	 * @return true or false depending on configuration status
	 */
	private IScenarioSetupHandler createHandler(String[] args) {

		options = buildOptions();

		// create the parser
		CommandLineParser parser = new GnuParser();
		
		try {
			// parse the command line arguments
			CommandLine line = parser.parse(options, args);
			if (line.hasOption(CLASS) == false
					&& line.hasOption("scenario") == false) {
				error("at least one of -scenario or -class must be specified");
				return null;

			} else {
				// all good so far get classes
				classestoload = new LinkedList();

				Option[] parms = line.getOptions();

				for (int i = 0; i < parms.length; i++) {
					Option parm = parms[i];
					String opt = parm.getOpt();
					if (opt.equals("scenario")) {
						String value = parm.getValue();
						String clazzname = "org.apache.kato.tck.scenario"
								+ value + ".ScenarioLauncher" + value;
						Class clazz = loadScenarioClass(clazzname);
						if (clazz == null) {
							error("unable to locate scenario class "
									+ clazzname + " to match scenario " + value);
							return null;
						}
						classestoload.add(clazz);

					} else if (opt.equals(CLASS)) {
						String value = parm.getValue();
						Class clazz = loadScenarioClass(value);
						if (clazz == null) {
							error("unable to locate class called " + value);
							return null;
						}
						classestoload.add(clazz);
					}
				}

				
				
				
				String mode = line.getOptionValue(MODE2);
				if (mode.equals("run")) {
					
					// 	do we have the dump trigger class?
					IDumpTrigger trigger=instantiateTriggerClass(line.getOptionValue(DUMPTRIGGER));
					// do we have a overide to the dump type?
					String dumpFormat = line.getOptionValue(DUMPTYPE);
					if(dumpFormat!=null) trigger.setDumpFormat(dumpFormat);
					
					if(trigger==null) return null; // error already reported
					
					handler=new ScenarioSetupHandler(trigger);
					
					
				}
				else if(mode.equals("config")) {
					ConfigurationExtractor x=new ConfigurationExtractor();
					
					if(line.hasOption(CONFIGDIR)) {
						String configdir=line.getOptionValue(CONFIGDIR);
						x.setConfigDirectory(configdir);
					}
					
					handler=x;
				}
				else {
					error("mode option is neither run or config");
					return null;
				}
				
				
				return handler;
				
								
			}

		} catch (Exception exp) {

			throw new TCKException(exp);
		}	

	}
	/**
	 * Instantiates the trigger class named
	 * If the provided trigger class name is null then
	 * the default trigger class for the JVM level is used
	 * 
	 * @param triggerName
	 * @return trigger instantiated
	 */
	private IDumpTrigger instantiateTriggerClass(String triggerName) {
		triggerName=getTriggerClassName(triggerName);
		if(triggerName==null) {
			error("cannot locate trigger class called "+triggerName);
			return null;
		}
		IDumpTrigger trigger=loadTriggerClass(triggerName);
		if(trigger==null) {
			error("cannot instantiate trigger class called "+triggerName);
			return null;
		}
		return trigger;
	}
	/**
	 * @param optionValue
	 * @return
	 */
	private String getTriggerClassName(String dumpClassName) {

		if(dumpClassName!=null && dumpClassName.trim().equals("")==false) return dumpClassName;
		
		return StandardDumpTrigger.class.getName();
		
		
	}
	/**
	 * @param optionValue
	 * @return
	 */
	private IDumpTrigger loadTriggerClass(String className) {
		
		Class c=loadScenarioClass(className);
		if(c==null) return null;
		
		try {
			return (IDumpTrigger) c.newInstance();
		} catch (InstantiationException e) {
			
			e.printStackTrace();
		} catch (IllegalAccessException e) {
			
			e.printStackTrace();
		}
		
		return null;
	
	}
	
}
