/*******************************************************************************
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package org.apache.kato.tck.results.reporter;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.Reader;
import java.net.URL;
import java.util.LinkedList;
import java.util.List;

import org.apache.kato.tck.harness.ConsoleOutputConstants;


/**
 * Class to scrape the tck test results from a text stream
 * 
 */
public class ConsoleOutputResultsReader implements ConsoleOutputConstants{

	private int lineNo=0;
	private int count=0;
	
	public   TCKResultMatrix getResultSet(InputStream in ) throws IOException {
		
		InputStreamReader reader=new InputStreamReader(in);
		return getResultSet(reader);
		
	}
	
	public  TCKResultMatrix getResultSet(Reader reader) throws IOException {
		
		BufferedReader br=new BufferedReader(reader);
		
		int findEyeCatcher=findEyeCatcher(br);
		if(findEyeCatcher<0) throw new IOException("TCK results eyecatcher not found");
		
		
		
		TCKResultMatrix matrix=new TCKResultMatrix(new FullSortedProvider());
		
		
		while(true) {
			// process results
			System.out.println("Eyecatcher on line "+lineNo);
			TCKResultSet set=getTCKResultSet(findEyeCatcher,br);
			System.out.println("Read "+set.getTestcaseCount()+" testcases");
			if(set!=null) matrix.addResultSet(set);
			// get next result
			findEyeCatcher=findEyeCatcher(br);
			if(findEyeCatcher<0) break; // finished
			
		}
		
		return matrix;
		
	}

	private TCKResultSet getTCKResultSet(int offset,BufferedReader br) throws IOException {
		
		
		
		// skip line of equals
		String line=readLine(offset,br);
		
		TestExecutionConfig config=new TestExecutionConfig();
		
		// read dump type
		line=readLine(offset,br);
		config.dumptype=line.substring(DUMPTYPE.length()).trim();
		
		// vendor
		line=readLine(offset,br);
		config.vendor=line.substring(VENDOR.length()).trim();
		
		// vmtype
		line=readLine(offset,br);
		config.jvm=line.substring(VMTYPE.length()).trim();
		
		// version					
		line=readLine(offset,br);
		config.version=line.substring(VERSION.length()).trim();
		
		line=readLine(offset,br);
		config.wordsize=line.substring(WORDSIZE.length()).trim();
		
		// testsuite
		line=readLine(offset,br);
		config.testsuite=line.substring(TESTSUITE.length()).trim();
		
		// os
		line=readLine(offset,br);
		config.os=line.substring(OS.length()).trim();

		// arch
		line=readLine(offset,br);
		config.arch=line.substring(ARCH.length()).trim();
		
		// build
		line=readLine(offset,br);
		config.build=line.substring(BUILD.length()).trim();
		
		// date
		line=readLine(offset,br);
		String date=line.substring(DATE.length()).trim();
		
		// skip line of equals
	    line=readLine(offset,br);
		
		TCKResultSet set=new TCKResultSet(config,date,null);
		
		report(config);
		
		loadResults(set,br,offset);
		
		return set;
	}

	
	private void report(TestExecutionConfig config) {
		System.out.print("Loading tests for config ");
		System.out.print(" arch="+config.arch);
		System.out.print(" build="+config.build);
		System.out.print(" dumptype="+config.dumptype);
		System.out.print(" jvm="+config.jvm);
		System.out.print(" os="+config.os);
		System.out.println("testsuite="+config.testsuite);
	}

	private  void loadResults(TCKResultSet set, BufferedReader br, int offset) throws IOException {
	
		while(true) {
		String line=readLine(offset,br);
		if(line.indexOf(ConsoleOutputConstants.TCK_TEST_SUITE_EXECUTION_ENDS)>=0) break; // done
		TestcaseResult result=null;
		if(line.startsWith(TCK_TEST_PASSED)) {
			TestData data=getTestData(line.substring(TCK_TEST_PASSED.length()));
			fillInData(data,br,offset);
			result=new TestcaseResultPass(data.className,data.methodname,0);
			result.setOutput(data.lines);
			
		}
		else if(line.startsWith(TCK_TEST_FAILED)) {
			TestData data=getTestData(line.substring(TCK_TEST_FAILED.length()));
			fillInData(data,br,offset);
			result=new TestcaseResultFailure(data.className,data.methodname,0,"fail",data.stack);
			result.setOutput(data.lines);
		}
		else if(line.startsWith(TCK_TEST_ERROR)) {
			TestData data=getTestData(line.substring(TCK_TEST_ERROR.length()));
			fillInData(data,br,offset);
			result=new TestcaseResultFailure(data.className,data.methodname,0,"error",data.stack);
			result.setOutput(data.lines);
		}
		else {
			System.out.println("ignoring "+line);
			continue;
		}
		
		if(result!=null) {
			count++;
			set.addTestcaseResult(result); 
			
		}
		
		}
	}

	
	private void fillInData(TestData data, BufferedReader br, int offset) throws IOException {
		// load any lines mentioned...
		if(data.outputLines>0) {
			 data.lines=new LinkedList();
			data.stack=new LinkedList();
			List target=data.lines;
			
			for(int i=0;i<data.outputLines;i++) {
				String line=readLine(offset, br);
				if(line.indexOf(STACK_TRACE_START)>=0) {
					target=data.stack;
				}
				else if(line.indexOf(STACK_TRACE_END)>=0) {
					target=data.lines;
				}
				else {
					target.add(line);
				}
			}
			
		}
		
	}

	private  TestData getTestData(String fullname) {
		
		int openBracket=fullname.indexOf("(");
		if(openBracket<0) return null;
		int closeBracket=fullname.indexOf(")");
		if(closeBracket<0) return null;
		
		TestData data=new TestData();
		data.methodname=fullname.substring(0,openBracket);
		data.className=fullname.substring(openBracket+1,closeBracket);
		try {
		data.outputLines=Integer.parseInt(fullname.substring(closeBracket+1).trim());
		} catch(NumberFormatException e) {
			;
		}
		
		
		return data;
	}

	private  String readLine(int offset, BufferedReader br) throws IOException {
		
		String line=br.readLine();
		if(line==null) throw new IOException("unexpected end of TCK results");
		lineNo++;
		if(line.length()>offset) return line.substring(offset);
		return line;
	}

	private  int findEyeCatcher(BufferedReader reader) throws IOException {
		
		while(true) {
			String line=reader.readLine();
			if(line==null) return -1;
			lineNo++;
			int offset=line.indexOf(TCK_TEST_SUITE_EXECUTION_BEGINS);
			if(offset>=0) return offset;
				
			
		}
	
		
	}

	public   TCKResultMatrix getResultSet(URL source) throws IOException {
		InputStream in=source.openStream();
		TCKResultMatrix result=getResultSet(in);
		in.close();
		System.out.println("count="+count);
		return result;
		
	}
	
	 class TestData {
		String className;
		String methodname;
		int outputLines=0;
		List lines=null;
		List stack=null;
		
	}
}
