/*******************************************************************************
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package org.apache.kato.tck.scenario142.javaruntime;

import org.apache.kato.tck.harness.scenario.ScenarioElement;




/**
 * Tests that monitors work correctly.
 * 
 * Tests with variations on the number of notify waiters, enter waiters and owners,
 * up to a maximum of 2 waiters of each type (there can be only one owner, of course).
 * 
 * TODO Add tests where threads own one monitor and wait on another.
 * 
 */
public class SetupJavaMonitor_ObjectMonitors extends ScenarioElement  {
	public static final String monitorNotifyThreadName = "NotifyThread";

	public static final String monitorOwnerThreadName = "OwnerThread";

	public 	static final String monitorEnterThreadName = "EnterThread";
	public static final String namePrefix = "TestMonitors-";

	/**
	 * A unique class to allow DTFJ to find the correct object.
	 * 
	 */
	public class MonitorClass {
		private String id;

		public MonitorClass(String id) {
			this.id = id;
		}
	}

	public static final String monitorClassName = MonitorClass.class.getName()
			.replace('.', '/');

	public static class MonitorThreadConfig {
		public MonitorThreadConfig(int waiters, int enters, int owners) {
			this.waiters = waiters;
			this.enters = enters;
			this.owners = owners;
			this.id = generatePrefix(waiters, enters, owners);
		}

		public void setMonitor(Object monitor) {
			this.monitor = monitor;
		}
		public Object getMonitor() {
			return monitor;
		}

		public String id;
		int waiters, enters, owners;
		public Object monitor;

	}

	public static int NUM_MONITORS = 9;

	public static MonitorThreadConfig[] monitorConfigs;
	{
		int i=0;
		monitorConfigs = new MonitorThreadConfig[NUM_MONITORS];
		
		monitorConfigs[i++] = new MonitorThreadConfig(2, 2, 1);
		monitorConfigs[i++] = new MonitorThreadConfig(0, 2, 1);
		monitorConfigs[i++] = new MonitorThreadConfig(2, 0, 1);
		monitorConfigs[i++] = new MonitorThreadConfig(2, 0, 0);
		monitorConfigs[i++] = new MonitorThreadConfig(1, 1, 1);
		monitorConfigs[i++] = new MonitorThreadConfig(1, 0, 1);
		monitorConfigs[i++] = new MonitorThreadConfig(1, 0, 0);
		monitorConfigs[i++] = new MonitorThreadConfig(0, 0, 1);
		monitorConfigs[i++] = new MonitorThreadConfig(0, 1, 1);
			
	}

	

	public MonitorThreadConfig findMonitorThreadConfig(String id) {
		for (int i = 0; i < monitorConfigs.length; i++) {
			MonitorThreadConfig current = monitorConfigs[i];

			if (current.id.equals(id)) {
				return current;
			}
		}

		return null;
	}

	
	
	public void configureMonitors() {
		for (int i = 0; i < monitorConfigs.length; i++) {
			createMonitor(monitorConfigs[i].waiters, monitorConfigs[i].enters,
					monitorConfigs[i].owners);
		}
	}

	private static String generatePrefix(int waiters, int enters, int owners) {
		return namePrefix + waiters + "-" + owners + "-" + enters;
	}

	public void createMonitor(int waiters, int enters, int owners) {
		String prefix = generatePrefix(waiters, enters, owners);

		Object monitor = new MonitorClass(prefix);

		// These rules are very important.
		if (waiters == 0 && enters == 0 && owners == 0) {
			throw new IllegalArgumentException(
					"Can't create monitor with no threads");
		}

		if (waiters < 0 || enters < 0 || owners < 0) {
			throw new IllegalArgumentException(
					"Numbers of threads must be 0 or above.");
		}

		if (enters > 0 && owners == 0) {
			throw new IllegalArgumentException(
					"Can't create enters threads with no owner");
		}

		if (owners > 1) {
			throw new IllegalArgumentException(
					"Can't create monitor with more than one owner");
		}

		// The ordering is significant.
		for (int i = 0; i < waiters; i++) {
			new MonitorNotifyThread(prefix + "-" + monitorNotifyThreadName
					+ "-" + i, monitor).start();
		}

		for (int i = 0; i < owners; i++) {
			new MonitorOwnerThread(prefix + "-" + monitorOwnerThreadName + "-"
					+ i, monitor).start();
		}

		for (int i = 0; i < enters; i++) {
			new MonitorEnterThread(prefix + "-" + monitorEnterThreadName + "-"
					+ i, monitor).start();
		}
	}

	public MonitorThreadConfig findMonitorThreadConfig(int waiters,
			int enters, int owners) {
		for (int i = 0; i < monitorConfigs.length; i++) {
			MonitorThreadConfig current = monitorConfigs[i];
	
			if ((current.waiters == waiters) && (current.enters == enters)
					&& (current.owners == owners)) {
				return current;
			}
		}
	
		return null;
	}

	public class MonitorNotifyThread extends Thread {
		private Object monitor;
		private volatile boolean almostReady = false;

		public MonitorNotifyThread(String name, Object monitor) {
			super(name);
			this.setDaemon(true); // Allow JVM to exit.
			this.monitor = monitor;
		}

		public void run() {
			synchronized (monitor) {
				synchronized (this) {
					almostReady = true;
					this.notify();
				}
				while (true) {
					try {
						monitor.wait();
					} catch (InterruptedException e) {
						// Keep waiting...
					}
				}
			}
		}

		/**
		 * Start thread, and waits until the thread is about to release the
		 * monitor.
		 */
		public void start() {
			super.start();
			synchronized (this) {
				try {
					while (!almostReady) {
						this.wait();
					}
				} catch (InterruptedException e) {
					e.printStackTrace();
				}
			}

		}

	}

	/**
	 * A thread that tries to own a monitor.
	 * 
	 */
	public class MonitorOwnerThread extends Thread {
		private Object monitor;
		private volatile boolean owned = false;

		public MonitorOwnerThread(String name, Object monitor) {
			super(name);
			this.setDaemon(true); // Allow JVM to exit.
			this.monitor = monitor;
		}

		public void run() {
			synchronized (monitor) {
				synchronized (this) {
					owned = true;
					this.notify();
				}
				while (true) {
					// We have either entered, or are entering.

					try {
						sleep(10000);
					} catch (InterruptedException e) {
						// TODO Auto-generated catch block
						e.printStackTrace();
					}
				}
			}
		}

		public void start() {
			super.start();
			synchronized (this) {
				try {
					while (!owned) {
						this.wait();
					}
				} catch (InterruptedException e) {
					e.printStackTrace();
				}
			}

		}
	}

	/**
	 * A thread that tries to enter a monitor.
	 * 
	 */
	public class MonitorEnterThread extends Thread {
		private Object monitor;
		private volatile boolean owned = false;

		public MonitorEnterThread(String name, Object monitor) {
			super(name);
			this.setDaemon(true); // Allow JVM to exit.
			this.monitor = monitor;
		}

		public void run() {
			synchronized (this) {
				owned = true;
				this.notify();
			}
			synchronized (monitor) {
				// This code should never be reached.
			}
		}

		public void start() {
			super.start();
	
			synchronized (this) {
	
				try {
					while (!owned) {
						this.wait();
					}
				} catch (InterruptedException e) {
					e.printStackTrace();
				}
			}
			try {
				sleep(1000);
			} catch (InterruptedException e1) {
				e1.printStackTrace();
			}
		}
	}


	



	
}
