/*******************************************************************************
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package org.apache.kato.tests.junit;

import java.util.Iterator;

import javax.tools.diagnostics.image.CorruptData;
import javax.tools.diagnostics.image.CorruptDataException;
import javax.tools.diagnostics.runtime.java.JavaMonitor;
import javax.tools.diagnostics.runtime.java.JavaThread;


public class JavaMonitorTest extends AbstractImageTestcase
{
	private JavaMonitor _monitor = null;
	
	public  JavaMonitor defaultJavaMonitor()
	{
		Iterator it = defaultJavaRuntime().getMonitors().iterator();
		assertNotNull(it);
		assertTrue(it.hasNext());
		return (JavaMonitor) it.next();
	}
	
	protected void setUp() throws Exception
	{
		_monitor = defaultJavaMonitor();
		super.setUp();
	}

	/**
	 * Test method for 'javax.tools.diagnostics.runtime.java.JavaMonitor.getObject()'
	 * 
	 * Ensures that no unexpected exceptions are thrown
	 */
	public void testGetObject()
	{
		_monitor.getObject();
	}

	/**
	 * Test method for 'javax.tools.diagnostics.runtime.java.JavaMonitor.getName()'
	 * 
	 * Ensures that a non-null name is returned
	 */
	public void testGetName()
	{
		try {
			String name = _monitor.getName();
			assertNotNull(name);
		} catch (CorruptDataException e) {
			//this is technically allowed
		}
	}

	/**
	 * Test method for 'javax.tools.diagnostics.runtime.java.JavaMonitor.getOwner()'
	 * 
	 * Ensures that no unexpected exceptions are thrown
	 */
	public void testGetOwner()
	{
		try {
			_monitor.getOwner();
		} catch (CorruptDataException e) {
			//this is technically allowed
		}
	}

	/**
	 * Test method for 'javax.tools.diagnostics.runtime.java.JavaMonitor.getEnterWaiters()'
	 * 
	 * Ensures that either an empty iterator is returned, or a non-empty iterator where all the 
	 * elements are JavaThread of CorruptData is returned.
	 */
	public void testGetEnterWaiters()
	{
		Iterator iter = _monitor.getEnterWaiters().iterator();
		assertNotNull(iter);
		while (iter.hasNext()) {
			Object element = iter.next();
			assertTrue((element instanceof JavaThread) || (element instanceof CorruptData));
		}
	}

	/**
	 * Test method for 'javax.tools.diagnostics.runtime.java.JavaMonitor.getNotifyWaiters()'
	 * 
	 * Ensures that either an empty iterator is returned, or a non-empty iterator where all the 
	 * elements are JavaThread of CorruptData is returned.
	 */
	public void testGetNotifyWaiters()
	{
		Iterator iter = _monitor.getNotifyWaiters().iterator();
		assertNotNull(iter);
		while (iter.hasNext()) {
			Object element = iter.next();
			assertTrue((element instanceof JavaThread) || (element instanceof CorruptData));
		}
	}
	
	/**
	 * Test method for 'javax.tools.diagnostics.runtime.java.JavaMonitor.getID()'
	 * 
	 * Ensures that the pointer we receive is non-null.
	 */
	public void testGetID()
	{
		assertNotNull(_monitor.getID());
	}

	/**
	 * Verify that the equals call doesn't throw
	 */
	public void testEquals()
	{
		try{
			assertTrue(_monitor.equals(_monitor));
		} catch (Throwable t) {
			assertTrue(false);
		}
	}
	
	/**
	 * Verify that hashCode() doesn't throw and returns non-zero (technically zero is ok but it will be 
	 * flagged here to ensure that we aren't doing anything bad to create the hashcode)
	 */
	public void testHashCode()
	{
		try {
			assertTrue(0 != _monitor.hashCode());
		} catch (Throwable t) {
			assertTrue(false);
		}
	}
	
	/**
	 * Ensure that all the monitors we find which have owners have valid owners.  
	 * 
	 * Note:  This test may have to be removed or further qualified in the future if we want to restrict this test suite to only deal
	 * with spec behaviour (which would allow getOwner() to throw CorruptDataException) or to test corner cases and past bugs in the
	 * (since this test suite currently addresses both needs).
	 */
	public void testAllMonitorOwners()
	{
		try {
			Iterator it = defaultJavaRuntime().getMonitors().iterator();
			
			while (it.hasNext()) {
				JavaMonitor monitor = (JavaMonitor) it.next();
				JavaThread owner = monitor.getOwner();
				
				if (null != owner) {
					assertNotNull(owner.getName());
				}
			}
		} catch (Throwable t) {
			assertTrue(false);
		}
	}
}
