/*******************************************************************************
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package org.apache.kato.tests.junit;

import java.util.Iterator;

import javax.tools.diagnostics.image.CorruptData;
import javax.tools.diagnostics.image.CorruptDataException;
import javax.tools.diagnostics.image.DataUnavailable;
import javax.tools.diagnostics.image.ImageAddressSpace;
import javax.tools.diagnostics.image.ImagePointer;
import javax.tools.diagnostics.image.ImageSection;
import javax.tools.diagnostics.image.MemoryAccessException;
import javax.tools.diagnostics.runtime.java.JavaHeap;
import javax.tools.diagnostics.runtime.java.JavaMonitor;
import javax.tools.diagnostics.runtime.java.JavaObject;
import javax.tools.diagnostics.runtime.java.JavaReference;
import javax.tools.diagnostics.runtime.java.JavaRuntime;
import javax.tools.diagnostics.runtime.java.JavaVMInitArgs;
import javax.tools.diagnostics.runtime.java.JavaVMOption;


public class JavaRuntimeTest extends AbstractImageTestcase
{
	private JavaRuntime _runtime = null;
	
	protected void setUp() throws Exception
	{
		_runtime = defaultJavaRuntime();
		super.setUp();
	}

	/**
	 * Test method for 'javax.tools.diagnostics.runtime.java.JavaRuntime.getJavaVM()'
	 * 
	 * Ensure that the pointer exists
	 */
	public void testGetJavaVM()
	{
		try {
			assertNotNull(_runtime.getJavaVM());
		} catch (CorruptDataException e) {
			//allowed by spec
		} catch (Exception e) {
			//no other exceptions should be thrown so catch it as a failure instead of an error
			fail();
		}
	}

	/**
	 * Test method for 'javax.tools.diagnostics.runtime.java.JavaRuntime.getJavaClassLoaders()'
	 * 
	 * Ensures that there is at least one class loader in the VM
	 */
	public void testGetJavaClassLoaders()
	{
		try {
			Iterator loaders = _runtime.getJavaClassLoaders().iterator();
			assertNotNull(loaders);
			assertTrue(loaders.hasNext());
		} catch (Exception e) {
			//no exceptions should be thrown so catch it as a failure instead of an error
			fail();
		}
	}

	/**
	 * Test method for 'javax.tools.diagnostics.runtime.java.JavaRuntime.getetObjectAtAddress()'
	 * 
	 * Ensures that every object in the heap is also accessible by address,
	 * and that addresses outside the heap or not pointing to an object start
	 * throw the expected exceptions.
	 */
	public void testGetObjectAtAddress()
	{
		JavaHeap heap = defaultJavaHeap();
		JavaRuntime runtime = defaultJavaRuntime();
		ImageAddressSpace addressSpace = defaultAddressSpace();
		ImagePointer address=null;
		ImagePointer unalignedAddress=null;
		Iterator heapSections = heap.getSections().iterator();


		//determine the heap start and end address
		long heapStartAddress = Long.MAX_VALUE;
		long heapEndAddress = 0;

		Object nextElement = null;

		ImageSection currentSection = null;
		long sectionStart = 0;

		while (heapSections.hasNext()) {
			nextElement = heapSections.next();
			if (nextElement instanceof ImageSection) {
				currentSection = (ImageSection) nextElement;
				sectionStart = currentSection.getBaseAddress().getAddress();
				if (sectionStart < heapStartAddress) {
					heapStartAddress = sectionStart; 
				}
				if (sectionStart + currentSection.getSize() > heapEndAddress) {
					heapEndAddress = sectionStart + currentSection.getSize(); 
				}

			}

		}

		//check that every object in the heap can be retrieved by address
		boolean exception=false;
		for (Iterator objects = heap.getObjects().iterator();objects.hasNext();) {
			Object potentialObject = objects.next();
			JavaObject object = null;
			if (potentialObject instanceof JavaObject) {
				object = (JavaObject)potentialObject;
			} else {
				continue;
			}
			address=(ImagePointer)object.getID();
			try {
				runtime.getObjectAtAddress(address).getJavaClass();
				if (unalignedAddress == null) {
					unalignedAddress = address.add(1);
				}
			} catch (Exception e) {
				e.printStackTrace();
				exception=true;
				break;
			}

		}
		assertFalse(exception);


		// Check an address BEFORE the start of the heap. For some JVMs, off-heap objects are supported
		exception=false;
		try {
			long addr1 = ((heapStartAddress-1000) & (Long.MAX_VALUE-7L));
			runtime.getObjectAtAddress(addressSpace.getPointer(addr1));
		} catch (IllegalArgumentException e) {
            e.printStackTrace();
			exception=true;
		} catch (Exception e) {
		}
		assertFalse(exception);

		// Check an address AFTER the first object but not aligned with an object boundary. 
		// unalignedAddress is expected to throw a IllegalArgumentException
		
		exception=false;
		try {
			runtime.getObjectAtAddress(unalignedAddress);
		} catch (IllegalArgumentException e) {
		// IllegalArgumentException is the expected behaviour
		} catch (CorruptDataException e) {
			exception=true;
			e.printStackTrace();
		} catch (MemoryAccessException e) {
			exception=true;
			e.printStackTrace();
		} catch (DataUnavailable e) {
			exception=true;
			e.printStackTrace();
		}
		assertFalse(exception);

	}


	/**
	 * Test method for 'javax.tools.diagnostics.runtime.java.JavaRuntime.getThreads()'
	 * 
	 * Ensures that there is at least one known thread in the runtime
	 */
	public void testGetThreads()
	{
		try {
			Iterator it = _runtime.getThreads().iterator();
			assertNotNull(it);
			assertTrue(it.hasNext());
		} catch (Exception e) {
			//no exceptions should be thrown so catch it as a failure instead of an error
			fail();
		}
	}

	/**
	 * Test method for 'javax.tools.diagnostics.runtime.java.JavaRuntime.getCompiledMethods()'
	 * 
	 * Ensures that the iterator is non-null
	 */
	public void testGetCompiledMethods()
	{
		try {
			Iterator it = _runtime.getCompiledMethods().iterator();
			assertNotNull(it);
		} catch (Exception e) {
			//no exceptions should be thrown so catch it as a failure instead of an error
			fail();
		}
	}

	/**
	 * Test method for 'javax.tools.diagnostics.runtime.java.JavaRuntime.getMonitors()'
	 * 
	 * Ensures that the iterator is non-null, has at least one element and it is a JavaMonitor object.
	 */
	public void testGetMonitors()
	{
		try {
			Iterator it = _runtime.getMonitors().iterator();
			assertNotNull(it);
			assertTrue(it.hasNext());
			assertTrue(it.next() instanceof JavaMonitor);
		} catch (Exception e) {
			//no exceptions should be thrown so catch it as a failure instead of an error
			fail();
		}
	}

	/**
	 * Test method for 'javax.tools.diagnostics.runtime.java.JavaRuntime.getHeaps()'
	 * 
	 * Ensures that the iterator is non-null, has at least one element and it is a JavaHeap object.
	 */
	public void testGetHeaps()
	{
		try {
			Iterator it = _runtime.getHeaps().iterator();
			assertNotNull(it);
			assertTrue(it.hasNext());
			assertTrue(it.next() instanceof JavaHeap);
		} catch (Exception e) {
			//no exceptions should be thrown so catch it as a failure instead of an error
			fail();
		}
	}

	/**
	 * Test method for 'javax.tools.diagnostics.runtime.java.JavaRuntime.getTraceBuffer(String, boolean)'
	 * 
	 * Ensure that something non-null comes back when we ask for the trace buffers
	 */
	public void testGetTraceBuffer()
	{
		try {
			assertNotNull(_runtime.getTraceBuffer("trace", false));
		} catch (CorruptDataException e) {
			//this is considered safe since it is in the spec
		} catch (Exception e) {
			//no other exceptions should be thrown
			fail();
		}
	}

	/**
	 * Test method for 'org.apache.kato.runtime.ManagedRuntime.getVersion()'
	 * 
	 * Makes sure that the full version is non-NULL.
	 * TODO:  Make this test more strict
	 */
	public void testGetVersion()
	{
		try {
			assertNotNull(_runtime.getVersion());
		} catch (CorruptDataException e) {
			//this is considered safe since it is in the spec
		} catch (Exception e) {
			//no other exceptions should be thrown
			fail();
		}
	}

	public void testGetJavaVMInitArgs() 
	{
		try {
			JavaVMInitArgs args = _runtime.getJavaVMInitArgs(); 
			assertNotNull(args);

			int version = args.getVersion();
			assertTrue((version == JavaVMInitArgs.JNI_VERSION_1_1) 
					|| (version == JavaVMInitArgs.JNI_VERSION_1_2)
					|| (version == JavaVMInitArgs.JNI_VERSION_1_4));

			Iterator options = args.getOptions().iterator();
			assertNotNull(options);
			
			// make sure there's at least one option (not strictly required)
			//assertTrue(options.hasNext());

			while (options.hasNext()) {
				JavaVMOption option = (JavaVMOption)options.next();
				assertNotNull(option.getOptionString());
				// just make sure getExtraInfo() doesn't throw an exception
				option.getExtraInfo();
			}
		} catch (CorruptDataException e) {
			// this is permitted by the spec
		} catch (DataUnavailable e) {
			// this is permitted by the spec
		} catch (Exception e) {
			//no other exceptions should be thrown
			fail();
		}
	}

	/**
	 * Verify that the equals call doesn't throw
	 */
	public void testEquals()
	{
		try{
			assertTrue(_runtime.equals(_runtime));
		} catch (Exception e) {
			fail();
		}
	}

	/**
	 * Verify that hashCode() doesn't throw and returns non-zero (technically zero is ok but it will be 
	 * flagged here to ensure that we aren't doing anything bad to create the hashcode)
	 */
	public void testHashCode()
	{
		try {
			assertTrue(0 != _runtime.hashCode());
		} catch (Exception e) {
			fail();
		}
	}

	/**
	 * Ensures that we can walk over every object in every heap in the VM.  Note that this may take a long time on large heaps!
	 */
	public void AtestWalkAllHeaps()
	{
		try {
			Iterator it = _runtime.getHeaps().iterator();
			assertNotNull(it);
			assertTrue(it.hasNext());
			while (it.hasNext()) {
				Object next = it.next();
				assertTrue(next instanceof JavaHeap);
				JavaHeap heap = (JavaHeap) next;
				Iterator objects = heap.getObjects().iterator();
				assertNotNull(objects);
				while (objects.hasNext()) {
					Object one = objects.next();
					assertTrue(one instanceof JavaObject);
				}
			}
		} catch (Exception e) {
			//no exceptions should be thrown from here
			e.printStackTrace();
			fail();
		}
	}

	/**
	 * Ensures that we get at least one root Reference object from the JavaRuntime 
	 * 
	 */
	public void testGetHeapRoots()
	{

		Iterator it = _runtime.getHeapRoots().iterator();
		assertNotNull(it);
		assertTrue(it.hasNext());
		Object first = it.next();
		assertTrue( first instanceof JavaReference || first instanceof CorruptData);

	}
}
