/*******************************************************************************
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package org.apache.kato.katoview.commands;

import java.util.HashMap;
import java.util.Vector;
import java.util.Stack;
import java.util.Iterator;

import javax.tools.diagnostics.image.Image;

import org.apache.kato.katoview.Output;
import org.apache.kato.katoview.commands.helpers.Utils;


public abstract class Command {
	
	private String name;
	private String desc;
	private String longDesc;
	protected Vector child_commands;
	protected Output out;
	
	protected Command(Output o, String aName, String aDesc, String aLongDesc){
		out = o;
		name = aName;
		desc = aDesc;
		
		// WARNING: assumptions on aLongDesc:
		//  - the only whitespace is ' ' and '\n'; no tabs
		//  - there are no single words longer than 80 - 12 = 68 characters
		//   - 12 is from nIndent + nWrapIndent + 1
		longDesc = aLongDesc;
	}
	
	public String getName(){
		return name;
	}
	
	public String getDesc(){
		return desc;
	}
	
	public Vector getChildCommands(){
		return child_commands;
	}
	
	public void printHelp(String indent, int descOffset, Stack currArgs, Stack prevArgs)
	{
		if (null == currArgs)
		{
			// the calling function just wants us to print our short name
			
			String spaces = "";
			
			int numSpaces = descOffset;
			numSpaces -= indent.length();
			numSpaces -= name.length();
			for (int i = 0; i < numSpaces; i++)
				spaces += " ";

			out.print(indent + name + spaces + desc + "\n");
			
			if (null != child_commands)
			{
				Iterator itCommand = child_commands.iterator();
				while (itCommand.hasNext())
				{
					Command command = (Command)itCommand.next();
					command.printHelp(indent + " ", descOffset, currArgs, prevArgs);
				}
			}
			return;
		}
		if (currArgs.isEmpty())
		{
			// out of arguments; we should either print our long description (if we're
			//  a leaf node in the command tree) or get our children to print their
			//  short descriptions

			String cmdName = getNameFromStack(prevArgs);
			
			if (null == child_commands) {	// we're a leaf node in the command tree
				if (null == longDesc) {
					out.error("<long descriptions not implemented for \"" +
							cmdName + "\" command>");
				} else {
					if (cmdName != null && cmdName != "") {
						out.print("\n");
						out.print("\t");
						out.print("\"" + cmdName + "\" command");
					}
					out.print("\n");
					out.print(formatDesc(longDesc));
					out.print("\n");
				}
			} else {	// we're an internal node in the command tree
				indent = "\t";
				out.print("\n");
				out.print(indent + cmdName);
				out.print("\n");
				Iterator itCommand = child_commands.iterator();
				while (itCommand.hasNext())
				{
					Command command = (Command)itCommand.next();
					command.printHelp(indent + " ", descOffset, null, prevArgs);
				}
				out.print("\n");
			}
			
			return;
		}
		
		// ASSERT: !currArgs.isEmpty()
		//  we've still got arguments left so we should print an error if we're a leaf
		//  node in the command tree or keep on processing if we're an internal node
		String arg = (String)currArgs.pop();
		
		if (null == child_commands)	// we're a leaf node; output an error
		{
			String cmdName = getNameFromStack(prevArgs);
			out.error("\"" + cmdName + "\" does not have any sub-commands");
			return;
		}
		
		// ASSERT: !currArgs.isEmpty() && null != child_commands
		//  we've still got arguments left and we're an internal node in the command tree;
		//  thus, we'll pass on the job to the next command down
		Iterator itCommand = child_commands.iterator();
		boolean found = false;
		Command currCommand = null;
		while (itCommand.hasNext() && !found)
		{
			currCommand = (Command)itCommand.next();
			if (currCommand.getName().equals(arg))
				found = true;
		}
		
		if (found)
		{
			prevArgs.push(arg);
			currCommand.printHelp(indent, descOffset, currArgs, prevArgs);
		}
		else
		{
			String cmdName = getNameFromStack(prevArgs);
			out.error("\"" + arg + "\" is not a sub-command of \"" + cmdName + "\"");
			return;
		}
	}
	
	private String getNameFromStack(Stack args)
	{
		String cmdName = "";
		while (!args.isEmpty())
		{
			String currArg = (String)args.pop();
			if (currArg.endsWith("/")) {
				cmdName = currArg + cmdName;
			} else {
				cmdName = currArg + " " + cmdName;
			}
		}
		if (cmdName.length() > 0 && cmdName.endsWith(" "))
		{
			cmdName = cmdName.substring(0, cmdName.length() - 1);
		}
		return cmdName;
	}
	
	private String formatDesc(String sDesc)
	{
		int nIndent = 8;
		
		// note: lines _always_ start with "\t " (the space is intentional) since
		//  every word is prefixed by a space
		String sIndent = "\t";
		
		int nWrapIndent = 3;
		String sWrapIndent = "   ";
		
		int nWidth = 80;
		char[] aDesc = sDesc.toCharArray();
		String formatted = "";

		formatted += sIndent;
		int charsInLine = nIndent;
		String currWord;
		int i = 0;
		
		while (i < aDesc.length)
		{
			currWord = "";
			for (;i < aDesc.length && '\n' != aDesc[i] && ' ' != aDesc[i]; i++)
			{
				currWord += aDesc[i];
			}
			// advance another character, past the '\n' or ' '; otherwise we get stuck
			//  in an infinite loop
			i++;
			
			int currWordLength = currWord.length() + 1;	// the 1 is for the added space
			if (currWordLength + charsInLine > nWidth)
			{
				// start a new line
				formatted += "\n";
				formatted += sIndent;
				formatted += sWrapIndent;
				charsInLine = nIndent + nWrapIndent;
			}

			// add the word to the current line
			formatted += " " + currWord;
			charsInLine += currWordLength;
			
			// if the character directly following the last word read is a new-line
			//  character, start a new line
			if (i <= aDesc.length && '\n' == aDesc[i-1])
			{
				// start a new line
				formatted += "\n";
				formatted += sIndent;
				charsInLine = nIndent;
			}
		}
		
		return formatted;
	}
	
	public void doCommand(Stack args, Image loadedImage, HashMap properties){
		boolean found = false;
		String nextCommand = null;
		
		if (!args.isEmpty())
		{
			nextCommand = (String)args.pop();
		}
		else
		{
			out.error("specified command requires additional subcommand");
			return;
		}
		nextCommand = nextCommand.toLowerCase();
		
		Command currCommand = null;
		Iterator itCommand = child_commands.iterator();
		while (itCommand.hasNext() && !found)
		{
			currCommand = (Command)itCommand.next();
			if (nextCommand.equals(currCommand.getName()))
				found = true;
		}
		
		if (found){
			currCommand.doCommand(args, loadedImage, properties);
		}
		else{
			out.error("command \"" + nextCommand + "\" does not exist; type \"help\" " +
					"for a list of available commands");
			return;
		}
	}
	
	protected void doScrollCommand(Image loadedImage, HashMap properties, boolean direction){
		RootCommand rootCommand = (RootCommand)properties.get(Utils.RootCommand_OBJECT);
		if (null == rootCommand){
			out.error("ERROR - No mapping found for RootCommand object");
			return;
		}
		
		Long currentMemAddress = (Long)properties.get(Utils.CURRENT_MEM_ADDRESS);
		Integer currentNumBytesToPrint = (Integer)properties.get(Utils.CURRENT_NUM_BYTES_TO_PRINT);
		if (null == currentMemAddress || null == currentNumBytesToPrint){
			out.error("ERROR - starting mem address or the number of bytes to print needs to be set " +
					"by running hexdump command");
			return;
		}
		
		//out.println("*******" + Long.toHexString( ((Long)properties.get(Utils.CURRENT_MEM_ADDRESS)).longValue() ));
		
		long newMemAddrss;
		if (direction)
			newMemAddrss = currentMemAddress.longValue() + currentNumBytesToPrint.intValue();
		else
			newMemAddrss = currentMemAddress.longValue() - currentNumBytesToPrint.intValue();
		
		Stack plusCommandStack = Utils.constructStackFromString("hexdump" + " " 
				+ Long.toHexString(newMemAddrss) + " " + currentNumBytesToPrint.toString());
		
		rootCommand.doCommand(plusCommandStack, loadedImage, properties);

		properties.put(Utils.CURRENT_MEM_ADDRESS, new Long(newMemAddrss));
	}
}

