/*******************************************************************************
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package org.apache.kato.katoview.commands;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Stack;

import javax.tools.diagnostics.image.CorruptDataException;
import javax.tools.diagnostics.image.Image;
import javax.tools.diagnostics.image.ImagePointer;
import javax.tools.diagnostics.image.ImageSection;
import javax.tools.diagnostics.image.MemoryAccessException;

import org.apache.kato.katoview.Output;
import org.apache.kato.katoview.commands.helpers.Utils;


public class FindCommand extends Command{
	
	static final int text = 1;
	static final int binary = 2;
	
	/*args for the find command must be comma separated
	 * e.g. find cpu,0,21665f64,64,16,30
	 */	
	FindAttribute findAtt = new FindAttribute();
	StringBuffer sb = new StringBuffer();
	ArrayList matches = new ArrayList();
	
	public FindCommand(Output o){
		super(o, "find", "searches memory for a given string",
				"parameters, comma separated: " + 
				"<pattern>,<start_address>,<end_address>,<memory_boundary>,<bytes_to_print>,<matches_to_display>\n\n" +
				"the find command searches for <pattern> in the memory segment from <start_address> to <end_address> inclusive, " +
				"matching only addresses that start at the specified <memory_boundary>, for example 1,2,4 or 8 bytes. It " +
				"displays <bytes_to_print> bytes from the final match, and lists the first <matches_to_display> matches found." +
				"\n example: find pattern,,,,64,3"
				);
		child_commands = null;
	}
	
	public void doCommand(Stack args, Image loadedImage, HashMap properties){
		String argsConcact = Utils.concatArgsFromStack(args);
		String[] params = argsConcact.split(",", -1);
		sb = new StringBuffer();
		matches.clear();
		if (!isParametersValid(args, params)) return;
		determineModeFromPattern();
		if (!parseParams(args, params)) return;
		Iterator imageSections = Utils.getAddressSapceSectionInfo(loadedImage);
		
		while(imageSections.hasNext()){
			if (matches.size() > findAtt.numMatchesToDisplay) break;
			ImageSection imageSection = (ImageSection)imageSections.next();				
			if (scanImageSection(imageSection)) break;
		}
		if (matches.size() > 0) 
			findAtt.lastMatch = ((Long)matches.get(matches.size()-1)).longValue();
		properties.put(Utils.FIND_ATTRIBUTES, findAtt);
		doPrint();
		if (matches.size() > 0) 
			printLastMatchContent(loadedImage, properties);
		
		//preppend "0x" back to pattern for findnext command 
		restoreHexPrefix();
	}

	private void restoreHexPrefix(){
		if (findAtt.mode == binary)
			findAtt.pattern = "0x" + findAtt.pattern; 
	}
	
	private void printLastMatchContent(Image loadedImage, HashMap properties){
		RootCommand rootCommand = (RootCommand)properties.get(Utils.RootCommand_OBJECT);
		if (null == rootCommand){
			out.error("ERROR - No mapping found for RootCommand object");
			return;
		}
		Stack hexdumpCommandStack = Utils.constructStackFromString("hexdump" + " " 
				+ Long.toHexString(findAtt.lastMatch) + " "
				+ findAtt.numBytesToPrint);
		rootCommand.doCommand(hexdumpCommandStack, loadedImage, properties);
	}
	
	private void doPrint(){
		int size = matches.size();
		if(0 == size){
			sb.append("No matches found.\n");
		}
		else{
			int limit = Math.min(findAtt.numMatchesToDisplay, size);
			for(int i = 0; i < limit; i++){
				long match = ((Long)matches.get(i)).longValue();
				sb.append("#" + i + ": " + "0x" + Long.toHexString(match) + "\n");
			}
		}
		out.print(new String(sb));
	}
	
	private boolean scanImageSection(ImageSection imageSection){
		long imageStartAddress = imageSection.getBaseAddress().getAddress();
		long imageEndAddress = imageStartAddress + imageSection.getSize() - 1;
		if (findAtt.startAddress < imageStartAddress && findAtt.endAddress < imageStartAddress){
			return false;
		}
		else if(findAtt.startAddress > imageEndAddress && findAtt.endAddress > imageEndAddress){
			return false;
		}
		else if(findAtt.startAddress >= imageStartAddress && findAtt.endAddress <= imageEndAddress){
			return scanRegion(findAtt.startAddress, findAtt.endAddress, imageSection);
		}
		else if(findAtt.startAddress <= imageStartAddress && findAtt.endAddress <= imageEndAddress 
				&& findAtt.endAddress >= imageStartAddress){
			return scanRegion(imageStartAddress, findAtt.endAddress, imageSection);
		}
		else if(findAtt.startAddress <= imageEndAddress && findAtt.startAddress >= imageStartAddress 
				&& findAtt.endAddress >= imageEndAddress){
			return scanRegion(findAtt.startAddress, imageEndAddress, imageSection);
		}
		else{
			return scanRegion(imageStartAddress, imageEndAddress, imageSection);
		}
	}
	
	private boolean scanRegion(long start, long end, ImageSection imageSection){
		ImagePointer imagePointer = imageSection.getBaseAddress();
		long i;
		if (0 != start%findAtt.boundary) {
			i = start - start%findAtt.boundary + findAtt.boundary;
		} else {
			i = start;
		}
		int patternLength = findAtt.length();
		byte[] bytes = findAtt.getBytes();
		
		for(; i <= end; i+=findAtt.boundary){
			int j;
			for(j = 0; j < patternLength; j++){
				byte oneByte = bytes[j];
				if (getByteFromImage(imagePointer, i+j) == oneByte){
					continue;
				}
				else {
					break;
				}
			}
			if (j >= patternLength){
				matches.add(new Long(i));
				if (matches.size() == findAtt.numMatchesToDisplay) 
					return true;
			}
		}
		return false;
	}
	
	private byte getByteFromImage(ImagePointer imagePointer, long address){
		long imagePointerBaseAddress = imagePointer.getAddress();
		try{
			return imagePointer.getByteAt(address-imagePointerBaseAddress);
		}catch(CorruptDataException cde){
			return 0;
		}catch(MemoryAccessException mae){
			out.error("MemoryAccessException occurred at 0x" + Long.toHexString(address));
			return 0;
		}
	}
	
	private boolean parseParams(Stack args, String[] params) {
		if (params[1].equals("")) {
			findAtt.startAddress = 0;
		} else {
			findAtt.startAddress = Utils.longFromString(params[1]).longValue();
		}
		
		if (params[2].equals("")) {
			//TODO how to convert to "unsigned" long
			findAtt.endAddress = Long.MAX_VALUE;
		} else {
			findAtt.endAddress = Utils.longFromString(params[2]).longValue();
		}
		
		if (findAtt.startAddress > findAtt.endAddress) {
			out.error("start address must not be greater than end address");
			printHelp("\t",10,new Stack(),args);
			return false;
		}
		
		if (params[3].equals("")) {
			findAtt.boundary = 1;
		} else {
			findAtt.boundary = Integer.parseInt(params[3]);
			if (findAtt.boundary <= 0) {
				out.error("memory boundary must be a positive non-zero value");
				printHelp("\t",10,new Stack(),args);
				return false;
			}
		}
		
		if (params[4].equals("")) {
			findAtt.numBytesToPrint = 256;
		} else {
			findAtt.numBytesToPrint = Integer.parseInt(params[4]);
			if (findAtt.numBytesToPrint < 0) {
				out.error("bytes to print must be a non-negative value");
				printHelp("\t",10,new Stack(),args);
				return false;
			}
		}
		
		if (params[5].equals("")) {
			findAtt.numMatchesToDisplay = 1;
		} else {
			findAtt.numMatchesToDisplay = Integer.parseInt(params[5]);
			if (findAtt.numMatchesToDisplay < 0) {
				out.error("matches to display must be a non-negative value");
				printHelp("\t",10,new Stack(),args);
				return false;
			}
		}
		return true;
	}
	
	private void determineModeFromPattern(){
		if (findAtt.pattern.startsWith("0x")) {
			findAtt.pattern = findAtt.pattern.substring(2);
			findAtt.mode = binary;
			allignBits();
		}
		else{
			findAtt.mode = text;
		}
	}
	
	private void allignBits(){
		int patternLength = findAtt.pattern.length();
		if (0 != patternLength%2){
			findAtt.pattern = "0" + findAtt.pattern;
		}
	}
	
	private boolean isParametersValid(Stack args, String[] params){
		if (6 != params.length){
			out.error("incorrect number of parameters");
			printHelp("\t",10,new Stack(),args);
			return false;
		}
		findAtt.pattern = params[0]; 
		if (findAtt.pattern.equals("")) {
			out.error("missing search pattern string");
			printHelp("\t",10,new Stack(),args);
			return false;
		}
		return true;
	}
	
	class FindAttribute{
		String pattern;
		long startAddress, endAddress;
		int boundary, numBytesToPrint, numMatchesToDisplay;
		long lastMatch;
		int mode;
		
		public int length() {
			if (mode == binary) {
				return pattern.length() / 2;
			} else {
				return pattern.length();
			}
		}
		
		public byte[] getBytes() {
			if (mode == text) {
				return pattern.getBytes();
			} else {
				byte[] result = new byte[length()];
				for (int i = 0; i < length(); i++) {
					result[i] = (byte)Integer.parseInt(pattern.substring(i*2, i*2+2), 16);
				}
				return result;
			}
		}
	}
}
