/*******************************************************************************
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package org.apache.kato.katoview.commands.infocommands;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Stack;

import javax.tools.diagnostics.image.CorruptDataException;
import javax.tools.diagnostics.image.DataUnavailable;
import javax.tools.diagnostics.image.Image;
import javax.tools.diagnostics.image.ImageSection;
import javax.tools.diagnostics.runtime.ManagedRuntime;
import javax.tools.diagnostics.runtime.java.JavaHeap;
import javax.tools.diagnostics.runtime.java.JavaObject;
import javax.tools.diagnostics.runtime.java.JavaRuntime;

import org.apache.kato.katoview.Output;
import org.apache.kato.katoview.commands.Command;
import org.apache.kato.katoview.commands.helpers.Utils;


public class InfoHeapCommand extends Command {
	
	public InfoHeapCommand(Output o){
		super(o, "heap", "displays information about Java heaps",
				"parameters: none, \"*\", or heap name\n\n" +
				" none prints:\n" +
				" - heap name\n" +
				" - heap section\n" +
				" \"*\" or heap name prints the following information\n" +
				" about all heaps or the specified heap, respectively:\n" +
				" - heap name\n" +
				" - (heap size and occupancy)\n" +
				" - heap sections\n" +
				"  - section name\n" +
				"  - section size\n" +
				"  - whether the section is shared\n" +
				"  - whether the section is executable\n" +
				"  - whether the section is read only\n"
				);
		child_commands = null;
	}
	public void doCommand(Stack args, Image loadedImage, HashMap properties){
		
		
		String param = null;
		StringBuffer paramsToString = new StringBuffer(); 
		if (!args.isEmpty()){
			for (int i = (args.size())-1; i>=0; i--){
				paramsToString.append(((String) args.get(i)) + " ");
			}
			if (paramsToString != null){
				param = paramsToString.toString();
				param = param.trim();
			}
		}
		
		ManagedRuntime mr;
		Iterator itRuntime = Utils.getRuntimes(loadedImage);
		int count = 1;

		out.print("\n");
		while (itRuntime.hasNext()) {
			mr = (ManagedRuntime)itRuntime.next();
			if (mr instanceof JavaRuntime)
			{
				out.print("\tRuntime #" + count);
				out.print("\n");
				
				if (args.isEmpty()){
					printHeapInfo(null,(JavaRuntime)mr, out);
					out.print("\nUse \"info heap *\" or \"info heap <heap_name>\" " +
							"for more information.\n");
				} else if (param.indexOf("*")>=0){
					printHeapInfo(param,(JavaRuntime)mr, out);
				} else {
					
					boolean foundHeap = searchForHeap(param, (JavaRuntime)mr, out);
					
					if (!foundHeap){
						out.error("Unable to locate heap: \"" + param +"\".");
						out.print("\tAvailable heaps: \n");
						printHeapInfo(null,(JavaRuntime)mr, out);
					}
				}
			
			}
			count++;
		}
	}
	private void printHeapInfo(String param, JavaRuntime runtime, Output out){
		
		Iterator itHeaps = runtime.getHeaps().iterator();
		int countheaps = 1;
		
		while (itHeaps.hasNext())
		{
			JavaHeap theHeap = (JavaHeap)itHeaps.next();
			out.print("\t Heap #" + countheaps + ":  " + theHeap.getName()+ "\n");
			if (param != null){
				printSectionInfo(theHeap, out);
			}
			countheaps++;
		}
	}
	private void printSectionInfo(JavaHeap theHeap, Output out){
		
		Iterator itSections = theHeap.getSections().iterator();
		int countSections = 1;
		
		while (itSections.hasNext()){
			
			ImageSection theSection = (ImageSection)itSections.next();
			out.print("\t  Section #"+ countSections + ":  " + theSection.getName() + "\n");
			out.print("\t   Size:        " + theSection.getSize() + " bytes\n");
			try 
			{
				out.print("\t   Shared:      " + theSection.isShared() + "\n");
				out.print("\t   Executable:  " + theSection.isExecutable() +"\n");
				out.print("\t   Read Only:   " + theSection.isReadOnly() + "\n");
			}
			catch(DataUnavailable e){
				out.print("\t   Shared:      <unknown>\n");
				out.print("\t   Executable:  <unknown>\n");
				out.print("\t   Read Only:   <unknown>\n");
			}
			countSections++;
		}
	}
	private boolean searchForHeap(String param, JavaRuntime jr, Output out){
		
		boolean foundHeap = false;
		
		Iterator itHeaps = jr.getHeaps().iterator();
		int countheaps = 1;
		
		while (itHeaps.hasNext())
		{
			JavaHeap theHeap = (JavaHeap)itHeaps.next();
			if (theHeap.getName().indexOf(param)==0){
				out.print("\t Heap #" + countheaps + ":  " + theHeap.getName()+"\n");
				printOccupancyInfo(theHeap, out);
				printSectionInfo(theHeap, out);
				foundHeap = true;
			}
			countheaps++;
		}
		
		return foundHeap;
	}

	private void printOccupancyInfo(JavaHeap theHeap, Output out){
		/*
		 * This method takes a lot of time and hence this information is only included 
		 * when using "info heap <heapname>". If this method was run for every heap 
		 * when using "info heap *" the amount of time it would take would be astronomical.
		 */
		long size = 0;
		long totalObjectSize = 0;
		
		Iterator itSections = theHeap.getSections().iterator();
		while (itSections.hasNext()){
			ImageSection theSection = (ImageSection)itSections.next();
			size = size + theSection.getSize();
		}
		out.print("\t  Size of heap: "+ size + " bytes\n");
		
		Iterator itObjects = theHeap.getObjects().iterator();
		try{
			while (itObjects.hasNext()){
				JavaObject theObject = (JavaObject)itObjects.next();
				totalObjectSize = totalObjectSize + theObject.getSize();
			}
			
			float percentage = ((float)totalObjectSize/(float)size)*10000; 
			int trimmedPercent = ((int)percentage); // Sending this float through an int gets it down to 2 decimal places.
			percentage = ((float)trimmedPercent)/100;
			
			out.print("\t  Occupancy :   "+ totalObjectSize + " bytes  (" + percentage + "%)\n");
		
		} catch (CorruptDataException e){
			out.print("\t  Occupancy :   <unknown>\n");
		}
		
	}
	
}
