/*******************************************************************************
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package org.apache.kato.katoview.commands.infocommands;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Stack;
import java.util.Properties;
import java.util.Enumeration;

import javax.tools.diagnostics.image.CorruptData;
import javax.tools.diagnostics.image.CorruptDataException;
import javax.tools.diagnostics.image.DataUnavailable;
import javax.tools.diagnostics.image.Image;
import javax.tools.diagnostics.image.ImageAddressSpace;
import javax.tools.diagnostics.image.ImageProcess;
import javax.tools.diagnostics.image.ImageThread;

import org.apache.kato.katoview.Output;
import org.apache.kato.katoview.commands.Command;
import org.apache.kato.katoview.commands.helpers.Exceptions;


public class InfoProcCommand extends Command {

	public InfoProcCommand(Output o){
		super(o, "proc", "displays threads, command line arguments, environment " +
				"variables, and shared modules of current process",
				
				"parameters: none\n\n" +
				"prints the following information about the current process\n\n" +
				"  - thread IDs for all its threads\n" +
				"  - the command line arguments it's using\n" +
				"  - its environment variables\n\n" +
				"note: to view the shared modules used by a process, use the " +
				"\"info sym\" command\n"
		);
		child_commands = null;
	}
	
	public void doCommand(Stack args, Image loadedImage, HashMap properties){
		if (args.isEmpty())
		{
			out.print("\n");
			printAddressSpaceInfo(loadedImage, out);
			out.print("\n");
		}
		else
		{
			out.error("\"info proc\" does not take any parameters");
			return;
		}
	}
	
	private void printAddressSpaceInfo(Image loadedImage, Output out)
	{
		Iterator itAddressSpace;
		ImageAddressSpace ias;
		int asnum = 0;
		
		itAddressSpace = loadedImage.getAddressSpaces().iterator();
		while (itAddressSpace.hasNext()) 
		{
			ias = (ImageAddressSpace)itAddressSpace.next();
			out.print("\taddress space # " + asnum + "\n");
			
			ImageProcess ip = ias.getCurrentProcess();
			if (ip == null) 
			{
				out.print("\t(No current process in this address space)\n");
			} 
			else 
			{
					printProcessInfo(ip, out);
			}
			asnum++;
		}
	}
	
	private void printProcessInfo(ImageProcess ip, Output out)
	{
		out.print("\n");
		printThreads(ip, out);
		out.print("\n");
		printCommandLine(ip, out);
		out.print("\n");
		printEnvironmentVariables(ip, out);
		out.print("\n");
	}
	
	private void printThreads(ImageProcess ip, Output out)
	{
		out.print("\t Native thread IDs for current process:");
		out.print("\n\t  ");
		int lineLength = 10; // normal tab plus 2 spaces

		// FIXME: should output architecture (32/64-bit), endianness before threads

		Iterator itThread = ip.getThreads().iterator();
		while (itThread.hasNext())
		{
			Object next = itThread.next();
			if (next instanceof CorruptData)
		        continue;
			ImageThread it = (ImageThread)next;

			try {
				String threadID = it.getID();
				if (threadID != null) {
					if (lineLength + threadID.length() > 80) {
						out.print("\n\t  ");
						lineLength = 10;
					}
					out.print(it.getID() + " ");
					lineLength += threadID.length() + 1;
				}
			} catch (CorruptDataException e) {
				out.print(Exceptions.getCorruptDataExceptionString());
			}
		}
	out.print("\n");
	}

	private void printCommandLine(ImageProcess ip, Output out)
	{
		out.print("\t Command line arguments used for current process:");
		out.print("\n");
		out.print("\t  ");
		
		try {
			String commandLine = ip.getCommandLine();
			if (null == commandLine)
				out.print("<null>");
			else
				out.print(commandLine);
		} catch (CorruptDataException e) {
			out.print(Exceptions.getCorruptDataExceptionString());
		} catch (DataUnavailable e) {
			out.print(Exceptions.getDataUnavailableString());
		}
		out.print("\n");
	}
	
	private void printEnvironmentVariables(ImageProcess ip, Output out)
	{
		out.print("\t Environment variables for current process:");
		out.print("\n");
		
		Properties variables;
		try {
			variables = ip.getEnvironment();
		} catch (CorruptDataException e) {
			out.print("\t  " + Exceptions.getCorruptDataExceptionString() + "\n");
			return;
		} catch (DataUnavailable e) {
			out.print("\t  " + Exceptions.getDataUnavailableString() + "\n");
			return;
		}
		
		Enumeration keys = variables.propertyNames();
		while (keys.hasMoreElements())
		{
			String key = (String)keys.nextElement();
			printVariableInfo(key, variables.getProperty(key), out);
		}
	}
	
	private void printVariableInfo(String key, String value, Output out)
	{
		out.print("\t  " + key + "=" + value + "\n");
	}
}
