/*******************************************************************************
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package org.apache.kato.katoview.commands.infocommands;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Properties;
import java.util.Stack;

import javax.tools.diagnostics.image.CorruptDataException;
import javax.tools.diagnostics.image.Image;
import javax.tools.diagnostics.image.DiagnosticException;
import javax.tools.diagnostics.image.MemoryAccessException;
import javax.tools.diagnostics.runtime.java.JavaClass;
import javax.tools.diagnostics.runtime.java.JavaField;
import javax.tools.diagnostics.runtime.java.JavaObject;
import javax.tools.diagnostics.runtime.java.JavaRuntime;

import org.apache.kato.katoview.Output;
import org.apache.kato.katoview.commands.Command;
import org.apache.kato.katoview.commands.helpers.JavaRuntimeVisitor;
import org.apache.kato.katoview.commands.helpers.KatoSearch;

public class InfoPropertyCommand extends Command {
	private static final String SHORT_DESCRIPTION = "displays system properties";
	private static final String COMMAND_NAME = "property";
	private static final String LONG_DESCRIPTION = 
		"Prints out all system properties or just properties passed as arguments.";

	public InfoPropertyCommand(Output o) {
		super(o, COMMAND_NAME, SHORT_DESCRIPTION, LONG_DESCRIPTION);
	}

	public void doCommand(Stack args, Image loadedImage, HashMap properties){
		try {
			KatoSearch.visitRuntimes(loadedImage, new PropertyGetter(args));
		} catch (DiagnosticException e) {
			e.printStackTrace();
		}
	}

	
	private class PropertyGetter implements JavaRuntimeVisitor {
		private Stack args;
		
		int runtimeNum = 0;
		public PropertyGetter(Stack args) {
			this.args = args;
		}

		public void visit(JavaRuntime runtime) throws DiagnosticException {
			int thisRuntime = runtimeNum++;
			
			JavaClass systemClass = KatoSearch.findClass(runtime, "java/lang/System");
			Properties properties = new Properties();
			
			if (systemClass != null) {
				try{
					JavaField props = KatoSearch.getField(systemClass, "props");
					if (props == null) {
						return;
					}
					JavaObject systemProperties = (JavaObject) props.get(null);
					if (systemProperties == null) {
						return;
					}
					JavaClass propertyClass = systemProperties.getJavaClass();
					JavaField tableField = KatoSearch.getField(propertyClass,"table");
					JavaObject table = (JavaObject) tableField.get(systemProperties);

					JavaClass entryClass = KatoSearch.findClass(runtime, "java/util/Hashtable$Entry");
					if (entryClass == null) {
						System.err.println("Couldn't find Entry class");
						return;
					}

					JavaField keyField = KatoSearch.getField(entryClass, "key");			
					if (keyField == null) {
						System.err.println("Couldn't find key field");
						return;
					}

					JavaField valueField = KatoSearch.getField(entryClass, "value");
					if (valueField == null) {
						System.err.println("Couldn't find value field");
						return;
					}

					JavaField nextField = KatoSearch.getField(entryClass, "next");
					if (nextField == null) {
						System.err.println("Couldn't find next field");
						return;
					}

					int tableSize = table.getArraySize();
					JavaObject[] entries = new JavaObject[tableSize];

					table.arraycopy(0, entries, 0, tableSize);

					for (int n=0; n < tableSize; n++) {
						JavaObject entry = entries[n];

						while (entry != null) {

							String key = keyField.getString(entry);
							String value = valueField.getString(entry);

							properties.setProperty(key, value);							
							entry = (JavaObject) nextField.get(entry);
						}
					}			

					out.println("Runtime #"+thisRuntime);
					if(args.size() == 0) {
						Iterator propertyEntries = properties.entrySet().iterator();
						
						while( propertyEntries.hasNext()) {
							Map.Entry prop = (Map.Entry) propertyEntries.next();
							System.out.println(
									((String) prop.getKey()) + "=" +
									((String) prop.getValue())							
									);
						}
					} else {						
						Iterator iter = args.iterator();
						while( iter.hasNext()) {
							String arg = (String) iter.next();							
							System.out.println(
									arg + "=" +
									properties.getProperty(arg)						
									);
						}
						
					}
				} catch (CorruptDataException e) {
					e.printStackTrace();
				} catch (MemoryAccessException e) {
					e.printStackTrace();
				}				
			}

		}
	}
}

