/*******************************************************************************
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package org.apache.kato.katoview.commands.setcommands;

import java.util.HashMap;
import java.util.Stack;

import javax.tools.diagnostics.image.Image;

import org.apache.kato.katoview.Output;
import org.apache.kato.katoview.commands.Command;
import org.apache.kato.katoview.heapdump.HeapDumpSettings;


public class SetHeapdumpCommand extends Command
{
	private static final String SHORT_DESCRIPTION = "configures heapdump format, filename and multiple heap support";
	private static final String COMMAND_NAME = "heapdump";
	private static final String LONG_DESCRIPTION = "parameters: [phd|txt], [file <filename>], [multiplefiles on|off]\n\n" 
		+ "[phd|txt] - the format for the heapdump. Default: phd.\n"
		+ "[file <filename>] - the file to write the heapdump to. Default: <core file name>.phd or <core file name>.txt.\n\n"
		+ "[multiplefiles on|off] - if set to on, multiple heaps are written to separate heapdumps. If set to off, multiple heaps are written " +
				"to the same heapdump. Default: off.\n\n"
		+ "Use \"show heapdump\" to see current settings.\n";

	public SetHeapdumpCommand(Output o)
	{
		super(o, COMMAND_NAME, SHORT_DESCRIPTION, LONG_DESCRIPTION);
	}

	public void doCommand(Stack args, Image loadedImage, HashMap properties)
	{
		if(args.isEmpty()) {
			out.error("\"set heapdump\" requires at least one parameter\n");
			return;
		}
		
		String arg1 = (String) args.pop();
		
		if(arg1.equalsIgnoreCase("phd")) {
			HeapDumpSettings.setPHDHeapDumps(properties);
			
			out.print("Heapdump format set to PHD\n");
		} else if (arg1.equalsIgnoreCase("txt")){
			HeapDumpSettings.setClassicHeapDumps(properties);
			
			out.print("Heapdump format set to classic (TXT)\n");
		} else if (arg1.equalsIgnoreCase("file")) {
			if(args.isEmpty()) {
				out.error("\"set heapdump file\" requires at least one parameter\n");
				return;
			}
			
			if(args.size() > 1) {
				out.error("\"set heapdump file\" accepts 1 parameter. You supplied " + args.size() + ".\n");
				return;
			}
			
			String filename = (String) args.pop();
			
			String originalFileName = HeapDumpSettings.getFileName(properties);
			
			HeapDumpSettings.setFileName(filename, properties);
			
			out.print("Heapdump file changed from " + originalFileName + " to " + filename + "\n");
		} else if (arg1.equalsIgnoreCase("multiplefiles")) {
			if(args.isEmpty()) {
				out.error("\"set heapdump multiplefiles\" requires one parameter: on or off\n");
				return;
			}
			
			if(args.size() > 1) {
				out.error("\"set heapdump multiplefiles\" requires one parameter: on or off. You suppled " + args.size() + " parameters\n");
				return;
			}
			
			String setting = (String) args.pop();
			
			if(setting.equalsIgnoreCase("on")) {
				out.println("Multiple heaps will be dumped into multiple heapdumps");
				HeapDumpSettings.setMultipleHeapsMultipleFiles(properties);
			} else if (setting.equalsIgnoreCase("off")) {
				out.println("Multiple heaps will be dumped into a single heapdump");
				HeapDumpSettings.setMultipleHeapsSingleFile(properties);
			} else {
				out.error("Unrecognised setting: " + setting + ". Valid options are \"on\" or \"off\"\n");
			}
		} else {
			out.error(arg1 + " is not a valid parameter for the \"set heapdump\" command");
		}
	}

}
