/*******************************************************************************
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package org.apache.kato.katoview.commands.setcommands;

import java.util.HashMap;
import java.util.Stack;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;

import javax.tools.diagnostics.image.Image;

import org.apache.kato.katoview.FileOutputChannel;
import org.apache.kato.katoview.Output;
import org.apache.kato.katoview.commands.Command;
import org.apache.kato.katoview.commands.helpers.Utils;


public class SetLoggingCommand extends Command {

	public SetLoggingCommand(Output o){
		super(o, "logging", "configures several logging-related parameters, " +
				"starts/stops logging",
				
				"parameters: [on|off], file <filename>, overwrite [on|off], " +
				"redirect [on|off]\n\n" +
				"- [on|off] - turns logging on or off (default: off)\n" +
				"- file <filename> - sets the file to log to; this will be relative " +
				"to the directory returned by the \"pwd\" command unless an " +
				"absolute path is specified; if the file is set while logging is on, " +
				"the change will take effect the next time logging is started " +
				"(default: <not set>)\n" +
				"- overwrite [on|off] - turns overwriting of the specified log file " +
				"on or off (off means that the log file will be appended to); if " +
				"this is on, the log file will be cleared every time \"set logging " +
				"on\" is run (default: off)\n" +
				"- redirect [on|off] - turns redirecting to file on or off (on means " +
				"that non-error output goes only to the log file (not the console) " +
				"when logging is on, off means that non-error output goes to both " +
				"the console and the log file); redirection must be turned off " +
				"logging can be turned off (default: off)\n"
		);
		child_commands = null;
	}
	
	public void doCommand(Stack args, Image loadedImage, HashMap properties){
		if (args.isEmpty())
		{
			out.error("\"set logging\" requires at least one parameter");
			return;
		}

		String arg1 = (String)args.pop();
		if (arg1.equals("on"))
		{
			if (!args.isEmpty())
			{
				out.error("\"set logging on\" does not take any additional parameters");
				return;
			}
			
			String logging = (String)properties.get("set_logging");
			if (null != logging && logging.equals("on"))
			{
				out.error("logging is already turned on");
				return;
			}
			
			String filename = (String)properties.get("set_logging_file");
			if (null == filename)
			{
				out.error("log file not yet set; set log file with \"set logging file <filename>\"");
				return;
			}
			
			File logfile = Utils.absPath(properties, filename);
			String filepath = logfile.getPath();
			FileWriter f;
			try {
				f = new FileWriter(logfile, appendToFile(properties));
			} catch (IOException e) {
				f = null;
				out.error("IOException encountered while opening file \"" + filepath +
						"\"; make sure the file can be written to");
				return;
			}

			out.addChannel(new FileOutputChannel(f), true);
			properties.put("set_logging", "on");
			properties.put("current_logging_file", filepath);
			
			out.print("\n");
			out.print("\tlogging turned on; outputting to \"" + filepath + "\"");
			out.print("\n\n");
		}
		else if (arg1.equals("off"))
		{
			if (!args.isEmpty())
			{
				out.error("\"set logging off\" does not take any additional parameters");
				return;
			}

			String redirect = (String)properties.get("set_logging_redirect");
			if (null != redirect && redirect.equals("on"))
			{
				out.error("cannot turn off logging when redirecting to only a file is " +
						"turned on; turn off redirecting to only a file and try again");
				return;
			}

			String logging = (String)properties.get("set_logging");
			if (null == logging || logging.equals("off"))
			{
				out.error("logging is already turned off");
				return;
			}
			
			properties.put("set_logging", "off");
			
			out.print("\n");
			out.print("\tlogging turned off; was logging to \"" +
					(String)properties.get("current_logging_file") + "\"");
			out.print("\n\n");
			
			out.removeFileChannel();
			properties.put("current_logging_file", null);
		}
		else if (arg1.equals("file"))
		{
			if (args.isEmpty())
			{
				out.error("\"set logging file\" requires a filename parameter");
				return;
			}

			String filename = (String)args.pop();
			
			if (!args.isEmpty())
			{
				out.error("\"set logging file\" takes exactly one parameter");
				return;
			}
			
			String oldFile = (String)properties.put("set_logging_file", filename);
			
			out.print("\n");
			if (null == oldFile) {
				out.print("\tlog file set to \"" + filename + "\"");
			} else {
				out.print("\tlog file changed from \"" + oldFile + "\" to \"" + filename + "\"");
			}
			out.print("\n\n");
		}
		else if (arg1.equals("overwrite"))
		{
			if (args.isEmpty())
			{
				out.error("\"set logging overwrite\" requires on/off parameter");
				return;
			}

			String param = (String)args.pop();
			
			if (!args.isEmpty())
			{
				out.error("\"set logging overwrite\" takes exactly one parameter");
				return;
			}
			
			if (param.equals("on"))
			{
				String oldParam;
				oldParam = (String)properties.put("set_logging_overwrite", "on");
				
				if (null == oldParam || oldParam.equals("off"))
				{
					out.print("\n");
					out.print("\toverwriting of log file option changed from \"off\" to \"on\"");
					out.print("\n\n");
				}
				else
				{
					out.print("\n");
					out.print("\toverwriting of log file option not changed (already \"on\")");
					out.print("\n\n");
				}
			}
			else if (param.equals("off"))
			{
				String oldParam;
				oldParam = (String)properties.put("set_logging_overwrite", "off");
				
				if (null == oldParam || oldParam.equals("off"))
				{
					out.print("\n");
					out.print("\toverwriting of log file option not changed (already \"off\")");
					out.print("\n\n");
				}
				else
				{
					out.print("\n");
					out.print("\toverwriting of log file option changed from \"on\" to \"off\"");
					out.print("\n\n");
				}
			}
			else
			{
				out.error("parameter for \"set logging overwrite\" must be \"on\" or \"off\"");
				return;
			}
		}
		else if (arg1.equals("redirect"))
		{
			if (args.isEmpty())
			{
				out.error("\"set logging redirect\" requires on/off parameter");
				return;
			}

			String param = (String)args.pop();
			
			if (!args.isEmpty())
			{
				out.error("\"set logging redirect\" takes exactly one parameter");
				return;
			}

			if (param.equals("on"))
			{
				String logging = (String)properties.get("set_logging");
				if (null == logging || logging.equals("off"))
				{
					out.error("cannot redirect to only a file because logging is off; " +
							"turn logging on and try again");
					return;
				}
				
				String oldParam;
				oldParam = (String)properties.put("set_logging_redirect", "on");
				
				if (null == oldParam || oldParam.equals("off"))
				{
					out.print("\n");
					out.print("\tredirecting to file only option changed from \"off\" to \"on\"");
					out.print("\n\n");
					out.setConsoleNoPrint(true);
				}
				else
				{
					out.print("\n");
					out.print("\tredirecting to file only option not changed (already \"on\")");
					out.print("\n\n");
				}
			}
			else if (param.equals("off"))
			{
				String oldParam;
				oldParam = (String)properties.put("set_logging_redirect", "off");
				
				if (null == oldParam || oldParam.equals("off"))
				{
					out.print("\n");
					out.print("\tredirecting to file only option not changed (already \"off\")");
					out.print("\n\n");
				}
				else
				{
					out.setConsoleNoPrint(false);
					out.print("\n");
					out.print("\tredirecting to file only option changed from \"on\" to \"off\"");
					out.print("\n\n");
				}
			}
			else
			{
				out.error("parameter for \"set logging redirect\" must be \"on\" or \"off\"");
				return;
			}
		}
		else
		{
			out.error("\"" + arg1 + "\" is not a valid parameter for the \"set logging\" command");
			return;
		}
	}
	
	private boolean appendToFile(HashMap properties)
	{
		String append = (String)properties.get("set_logging_overwrite");
		if (append == null || !append.equals("on"))
			return true;
		else
			return false;
	}

}
