/*******************************************************************************
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package org.apache.kato.katoview.commands.tomcat;

import java.util.HashMap;
import java.util.Stack;

import javax.tools.diagnostics.image.Image;
import javax.tools.diagnostics.image.DiagnosticException;
import javax.tools.diagnostics.runtime.java.JavaClass;
import javax.tools.diagnostics.runtime.java.JavaField;
import javax.tools.diagnostics.runtime.java.JavaRuntime;

import org.apache.kato.katoview.Output;
import org.apache.kato.katoview.commands.Command;
import org.apache.kato.katoview.commands.helpers.JavaClassVisitor;
import org.apache.kato.katoview.commands.helpers.JavaRuntimeVisitor;
import org.apache.kato.katoview.commands.helpers.KatoSearch;

/**
 * Retrieves the version of Tomcat in the dump, or displays an error.
 *
 */
public class TomcatVersionCommand extends Command {
	private static final String SHORT_DESCRIPTION = "Returns version of Tomcat in the dump.";
	private static final String COMMAND_NAME = "version";
	private static final String LONG_DESCRIPTION = "Looks for Tomcat within the dump and reports the version.";
	
	// Look for this class.
	private static final String SERVER_INFO_CLASS="org/apache/catalina/util/ServerInfo"; 
	private static final String SERVER_NUMBER = "serverNumber",
								SERVER_BUILT = "serverBuilt",
								SERVER_INFO = "serverInfo";
	
	public TomcatVersionCommand(Output o) {
		super(o, COMMAND_NAME, SHORT_DESCRIPTION, LONG_DESCRIPTION);
		
		child_commands = null;
	}

	public void doCommand(Stack args, Image loadedImage, HashMap properties) {
		try {
			KatoSearch.visitRuntimes(loadedImage, new PrintVersion());
		} catch (DiagnosticException e) {
			e.printStackTrace();
		}
	}

	/**
	 * A visitor for finding the correct class and then return
	 *
	 */
	private class PrintVersion implements JavaRuntimeVisitor, JavaClassVisitor {
		boolean found = false;
		
		public void visit(JavaRuntime runtime) throws DiagnosticException {
			found = false;
			KatoSearch.visitClasses(runtime, this);
			
			if (found == false) {
				out.println("ERROR: Couldn't find a Tomcat server");				
			}
		}

		public boolean visit(JavaClass clazz) throws DiagnosticException {
			if (clazz.getName().equals(SERVER_INFO_CLASS)) {
				JavaField serverNumberField = KatoSearch.getField(clazz, SERVER_NUMBER);
				JavaField serverBuiltField = KatoSearch.getField(clazz, SERVER_BUILT);
				JavaField serverInfoField = KatoSearch.getField(clazz, SERVER_INFO);
				
				
				out.println("Found Tomcat server: ");
				if (serverInfoField != null) {
					out.println("\tInfo: "+serverInfoField.getString(null));						
				}
				if (serverBuiltField != null) {
					out.println("\tBuilt: "+serverBuiltField.getString(null));						
				}
				if (serverNumberField != null) {
					out.println("\tNumber: "+serverNumberField.getString(null));						
				}
				found = true; // indicate we found a server.
				
				return false; // Stop searching through classes.
			}
			return true;
		}
		
	}
}